# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import sys
import typing as t
from datetime import date

from ..abstract import AbstractFile
from ..commons import ASCII, BLOCK_SIZE, IMAGE, READ_FILE_FULL
from ..device.abstract import AbstractDevice
from ..device.block_12bit import BlockDevice12Bit, COSRXBlockDevice12Bit
from .os8fs import OS8DirectoryEntry, OS8Filesystem, OS8Segment, os8_split_fullname

__all__ = [
    "COS300DirectoryEntry",
    "COS300Filesystem",
]

# COS 300/310 System Reference Manual, 1975, Pag 322
# https://bitsavers.org/pdf/dec/pdp8/cos-300/DEC-08-OCOSA-F_D_COS_300_310_System_Reference_Manual_Jul75.pdf

# There are 4 types of files in COS: Source, Binary, Data, and System files
COS_EXTENSIONS = {  # OS/8 extension: COS 1 character file type
    "DA": "A",  # Data
    "DB": "B",  # Binary
    "AS": "S",  # Source
    "SV": "V",  # System files (OS/8 SAVE Format)
}
# fmt: off
# COS 300/310 System Reference Manual, 1975, Pag 263
# https://bitsavers.org/pdf/dec/pdp8/cos-300/DEC-08-OCOSA-F_D_COS_300_310_System_Reference_Manual_Jul75.pdf
# In both source and data files, characters are stored 2 characters per word in 6-bit binary form
COS_CODES = [
    "\0", " ", "!", '"', "#", "$", "%", "&", "'", "(", ")", "*", "+", ",", "-", ".", "/",
    "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ":", ";", "<", "=", ">", "?",
    "@", "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O",
    "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\t", "]", "^"
]
# fmt: on
COS_INDEX = {ch: i for i, ch in enumerate(COS_CODES) if ch}


def split_ext(fullname: str) -> t.Tuple[str, str]:
    """
    Split a fullname into the filename and extension.
    """
    fullname = fullname.upper()
    try:
        filename, extension = fullname.split(".", 1)
        return filename, extension
    except Exception:
        return (fullname, "")


def cos_file_type_to_os8_extension(file_type: str) -> str:
    """
    Convert a COS file type to an OS/8 extension.
    """
    return {v: k for k, v in COS_EXTENSIONS.items()}.get(file_type, file_type)


def from_bytes_to_12bit_words(byte_data: bytes) -> t.List[int]:
    """
    Convert bytes to 12-bit words.
    """
    words = []
    for i in range(0, len(byte_data), 3):
        chr1 = byte_data[i] & 0xFF
        try:
            chr2 = byte_data[i + 1] & 0xFF
        except IndexError:
            chr2 = 0
        try:
            chr3 = byte_data[i + 2] & 0xFF
        except IndexError:
            chr3 = 0
        words.append(chr1 | ((chr3 & 0o360) << 4))
        words.append(chr2 | ((chr3 & 0o17) << 8))
    return words


def from_12bit_words_to_bytes(words: t.List[int]) -> bytes:
    """
    Convert 12-bit words to bytes.
    """
    result = bytearray()
    for i in range(0, len(words), 2):
        chr1 = words[i]
        try:
            chr2 = words[i + 1]
        except IndexError:
            chr2 = 0
        chr3 = ((chr2 >> 8) & 0o17) | ((chr1 >> 4) & 0o360)
        result.append(chr1 & 0xFF)
        result.append(chr2 & 0xFF)
        result.append(chr3 & 0xFF)
    return bytes(result)


def cos_codes_to_ascii(words: t.List[int]) -> bytes:
    """
    Source/data file format:

    +-----------------+
    |   Word count    |  1 word
    +-----------------+
    |   Line number   |  1 word
    +-----------------+
    |   Line data     |  n-1 words
    /                 |
    |                 |
    +-----------------+

    """
    out = bytearray()
    words = list(words)  # Make a copy of the list to avoid modifying the original
    while words:
        # Read the word count
        len_word = 4096 - words.pop(0) - 1
        if len_word >= 4095:
            break
        # Read the line number
        line_number = words.pop(0)
        out += f"{line_number:04} ".encode("ascii")  # Line number
        # Read the line
        for i in range(0, len_word):
            # Characters are stored 2 characters per word in 6-bit binary form
            b3 = words.pop(0)
            b1 = b3 >> 6
            b2 = b3 & 0o77
            out += COS_CODES[b1].encode("ascii")
            if b2 != 0:
                out += COS_CODES[b2].encode("ascii")
        out += b"\n"
    return bytes(out)


def ascii_to_cos_codes(text: bytes) -> t.List[int]:
    words: t.List[int] = []
    lines = text.decode("ascii").splitlines()

    for line in lines:
        if len(line) < 5 or not line[:4].isdigit():
            continue  # Skip invalid or malformed lines

        line_number = int(line[:4])
        content = line[5:]  # Skip past line number and space
        word_count = math.ceil(len(content) / 2)
        words.append(4096 - word_count - 1)  # Inverse of encoding logic
        words.append(line_number)

        for i in range(0, len(content), 2):
            c1 = content[i]
            b1 = COS_INDEX.get(c1, 0)
            if i + 1 < len(content):
                c2 = content[i + 1] if i + 1 < len(content) else " "
                b2 = COS_INDEX.get(c2, 0)
            else:
                b2 = 0
            word = (b1 << 6) | b2
            words.append(word)

    return words


class COS300DirectoryEntry(OS8DirectoryEntry):

    @property
    def file_type(self) -> t.Optional[str]:
        """File type"""
        if self.is_tentative:
            return "TEMP"
        elif self.is_empty:
            return "EMPTY"
        elif self.extension == "A":
            return "DATA"
        elif self.extension == "B":
            return "BINARY"
        elif self.extension == "S":
            return "SOURCE"
        elif self.extension == "V":
            return "SYSTEM"
        else:
            return "PERM"

    @classmethod
    def read(cls, segment: "OS8Segment", words: t.List[int], position: int, file_position: int) -> "OS8DirectoryEntry":
        self = super().read(segment, words, position, file_position)
        # Convert the extension to a COS 1 character extension
        self.extension = COS_EXTENSIONS.get(self.extension, self.extension)
        return self

    def to_words(self) -> t.List[int]:
        """
        Write the directory entry
        """
        try:
            # Convert the 1 character COS file type to an OS/8 extension
            tmp = self.extension
            self.extension = cos_file_type_to_os8_extension(self.extension)
            return super().to_words()
        finally:
            self.extension = tmp

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        is_source_file = self.extension == "S"  # Source file
        if file_mode is None:
            file_mode = ASCII if is_source_file else IMAGE
        # Always read the file as IMAGE
        f = self.open(IMAGE)
        try:
            data = f.read_block(0, READ_FILE_FULL)
            if file_mode == ASCII and is_source_file:
                words = from_bytes_to_12bit_words(data)
                return cos_codes_to_ascii(words)
            else:
                return data
        finally:
            f.close()

    def __str__(self) -> str:
        return f"{self.fullname:<10} {self.file_type:<6} {self.creation_date or '          '} {self.length:>6} {self.file_position:6d}"


class COS300Filesystem(OS8Filesystem):
    """
    COS-300/COS-310 Filesystem

    The COS-300/COS-310 filesystem is essentially the same filesystem as the OS/8.
    The main difference is that the COS-300/COS-310 filesystem has four types of files:
    - Source (.S)
    - Binary (.B)
    - Data (.A)
    - System files (.V)
    The file type is stored in the extension field of the directory entry.
    On the directory entry the file type is represented by two characters, e.g. "SV" for a system file,
    but on the extension is repsented by a single character, e.g. "V" for a system file.

    https://bitsavers.org/pdf/dec/pdp8/cos-300/DEC-08-OCOSA-F_D_COS_300_310_System_Reference_Manual_Jul75.pdf
    """

    fs_name = "cos300"
    fs_description = "PDP-8 COS-300/COS-310"
    fs_platforms = ["pdp-8"]
    directory_entry_class: t.Type[OS8DirectoryEntry] = COS300DirectoryEntry

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = COSRXBlockDevice12Bit(file_or_device)
        elif isinstance(file_or_device, BlockDevice12Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        device_type: t.Union[bool, str] = "",
        **kwargs: t.Union[bool, str],
    ) -> "COS300Filesystem":
        """
        Mount the COS-300/COS-310 filesystem from a file
        """
        self = cls(file_or_dev)
        self.current_partition = 0
        self.number_of_blocks = self.dev.get_size() // BLOCK_SIZE
        if strict:
            # Check if the filesystem is valid by reading the directory entries
            self.check_os8()
        return self

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        partition = self.current_partition
        if pattern:
            partition, pattern = os8_split_fullname(partition, pattern, wildcard=True)
        part = self.get_partition(partition)
        if not options.get("brief"):
            dt = date.today().strftime("%d-%b-%y").upper()
            sys.stdout.write(f"DIRECTORY       {dt}\n\n")
            sys.stdout.write("NAME   TYPE LN    DATE\n\n")
        for segment in part.read_dir_segments():
            for x in segment.filter_entries_list(pattern):
                fullname = f"{x.filename:<6}.{x.extension:<2}"
                if options.get("brief"):
                    sys.stdout.write(f"{fullname}\n")
                else:
                    file_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or ""
                    sys.stdout.write(f"{x.filename:<6}  {x.extension:>2}  {x.length:>02}  {file_date:<9}\n")
        if not options.get("brief"):
            unused = part.free()
            sys.stdout.write(f" <{unused:>04} FREE BLOCKS>\n")

    def read_bytes(self, fullname: str, file_mode: t.Optional[str] = None) -> bytes:
        """
        Get the content of a file
        """
        is_source_file = split_ext(fullname)[1] == "S"  # Source file
        if file_mode is None:
            file_mode = ASCII if is_source_file else IMAGE
        # Always read the file as IMAGE
        f = self.open_file(fullname, IMAGE)
        try:
            data = f.read_block(0, READ_FILE_FULL)
            if file_mode == ASCII and is_source_file:
                words = from_bytes_to_12bit_words(data)
                return cos_codes_to_ascii(words)
            else:
                return data
        finally:
            f.close()

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        is_source_file = split_ext(fullname)[1] == "S"  # Source file
        if file_mode is None:
            file_mode = ASCII if is_source_file else IMAGE
        if file_mode == ASCII and is_source_file:
            # Convert ASCII content to COS codes
            words = ascii_to_cos_codes(content)
            content = from_12bit_words_to_bytes(words)
            # Always write the file as IMAGE
            file_mode = IMAGE
        super().write_bytes(fullname, content, creation_date, file_type, file_mode)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import sys
import typing as t
from abc import ABC, abstractmethod
from datetime import date

from .commons import ASCII, BLOCK_SIZE, IMAGE, READ_FILE_FULL, hex_dump
from .device.abstract import AbstractDevice
from .device.block import BlockDevice

__all__ = [
    "AbstractFile",
    "AbstractDirectoryEntry",
    "AbstractFilesystem",
    "AbstractBlockFilesystem",
]


class AbstractFile(ABC):
    """Abstract base class for file operations"""

    current_position: int = 0

    @abstractmethod
    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """Read block(s) of data from the file"""

    @abstractmethod
    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """Write block(s) of data to the file"""

    @abstractmethod
    def get_size(self) -> int:
        """Get file size in bytes."""

    @abstractmethod
    def get_block_size(self) -> int:
        """Get file block size in bytes"""

    @abstractmethod
    def close(self) -> None:
        """Close the file"""

    def read(self, size: t.Optional[int] = None) -> bytes:
        """Read bytes from the file, starting at the current position"""
        data = bytearray()
        block_size = self.get_block_size()
        while size is None or len(data) < size:
            # Calculate current block and offset within the block
            block_number = self.current_position // block_size
            block_offset = self.current_position % block_size
            # print(f"{block_number=} {block_offset=}")
            # Read the next block
            block_data = self.read_block(block_number)
            # No more data
            if not block_data:
                break
            # Calculate the data to append
            if size is None:
                data_to_append = block_data[block_offset:]
            else:
                remaining_size = size - len(data)
                data_to_append = block_data[block_offset : block_offset + remaining_size]
            data.extend(data_to_append)
            self.current_position += len(data_to_append)
            # If size is specified and we have read enough data, break the loop
            if (size is not None and len(data) >= size) or (len(data_to_append) == 0):
                break
        return bytes(data)

    def write(self, data: bytes) -> int:
        """Write bytes to the file at the current position"""
        data_length = len(data)
        written = 0
        block_size = self.get_block_size()
        while written < data_length:
            # Calculate current block and offset within the block
            block_number = self.current_position // block_size
            block_offset = self.current_position % block_size
            # Read the current block
            block_data = bytearray(self.read_block(block_number))
            if not block_data:
                block_data = bytearray(block_size)
            # Calculate the amount of data to write in the current block
            remaining_block_space = block_size - block_offset
            data_to_write = data[written : written + remaining_block_space]
            # Write the data to the block
            block_data[block_offset : block_offset + len(data_to_write)] = data_to_write
            # Write the block to the device
            self.write_block(block_data, block_number)
            # Update current position and written count
            self.current_position += len(data_to_write)
            written += len(data_to_write)
        return written

    def seek(self, offset: int, whence: int = 0) -> None:
        """Move the current position in the file to a new location"""
        if whence == os.SEEK_SET:  # Absolute file positioning
            self.current_position = offset
        elif whence == os.SEEK_CUR:  # Seek relative to the current position
            self.current_position += offset
        elif whence == os.SEEK_END:  # Seek relative to the file's end
            self.current_position = self.get_size() + offset
        else:
            raise ValueError
        # Ensure the current position is not negative
        if self.current_position < 0:
            self.current_position = 0

    def tell(self) -> int:
        """Get current file position"""
        return self.current_position

    def truncate(self, size: t.Optional[int] = None) -> None:
        """
        Resize the file to the given number of bytes.
        If the size is not specified, the current position will be used.
        """
        if size is not None and self.current_position > size:
            self.current_position = size


class AbstractDirectoryEntry(ABC):

    @property
    @abstractmethod
    def fullname(self) -> str:
        """Name with path"""

    @property
    @abstractmethod
    def basename(self) -> str:
        """Final path component"""

    @property
    def creation_date(self) -> t.Optional[date]:
        """Creation date"""
        return None

    @property
    def file_type(self) -> t.Optional[str]:
        """File type"""
        return None

    @abstractmethod
    def get_length(self) -> int:
        """Get the length in blocks"""

    @abstractmethod
    def get_size(self) -> int:
        """Get file size in bytes"""

    @abstractmethod
    def get_block_size(self) -> int:
        """Get file block size in bytes"""

    @abstractmethod
    def delete(self) -> bool:
        """Delete the directory entry"""

    @abstractmethod
    def write(self) -> bool:
        """Write the directory entry"""

    @abstractmethod
    def open(self, file_mode: t.Optional[str] = None) -> AbstractFile:
        """Open the file"""

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        f = self.open(file_mode)
        try:
            return f.read_block(0, READ_FILE_FULL)[: f.get_size()]
        finally:
            f.close()

    def read_text(self, encoding: str = "ascii", errors: str = "ignore", file_mode: str = ASCII) -> str:
        """Get the content of the file as text"""
        data = self.read_bytes(file_mode)
        return data.decode(encoding, errors)


class AbstractFilesystem:
    """Abstract base class for filesystem implementations"""

    fs_name: str  # Filesystem name
    fs_description: str  # Filesystem description
    dev: AbstractDevice

    def __init__(self, dev: "AbstractDevice"):
        self.dev = dev

    @classmethod
    @abstractmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "AbstractFilesystem":
        """Mount the filesystem"""
        pass

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "AbstractFilesystem":
        """Initialize the filesystem"""
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    @abstractmethod
    def filter_entries_list(
        self, pattern: t.Optional[str], include_all: bool = False, expand: bool = True
    ) -> t.Iterator["AbstractDirectoryEntry"]:
        """Filter directory entries based on a pattern"""

    @property
    @abstractmethod
    def entries_list(self) -> t.Iterator["AbstractDirectoryEntry"]:
        """Property to get an iterator of directory entries"""

    @abstractmethod
    def get_file_entry(self, fullname: str) -> "AbstractDirectoryEntry":
        """Get the directory entry for a file"""

    @abstractmethod
    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """Write content to a file"""

    @abstractmethod
    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
    ) -> t.Optional["AbstractDirectoryEntry"]:
        """Create a new file with a given length in number of blocks"""

    def create_directory(
        self,
        fullname: str,
        options: t.Dict[str, t.Union[str, bool]],
    ) -> t.Optional["AbstractDirectoryEntry"]:
        """Create a new directory"""
        raise OSError(errno.ENOSYS, os.strerror(errno.ENOSYS))

    def chdir(self, fullname: str) -> bool:
        """Change the current directory"""
        return False

    def get_pwd(self) -> str:
        """Get the current directory"""
        return ""

    def isdir(self, fullname: str) -> bool:
        """Check if the given path is a directory"""
        return False

    @abstractmethod
    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        """List directory contents"""

    @abstractmethod
    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        """Examine the filesystem"""

    @abstractmethod
    def get_size(self) -> int:
        """Get filesystem size in bytes"""

    def close(self) -> None:
        """Close the filesystem"""
        self.dev.close()

    def exists(self, fullname: str) -> bool:
        """Check if the given path exists"""
        try:
            self.get_file_entry(fullname)
            return True
        except FileNotFoundError:
            return False

    def open_file(self, fullname: str, file_mode: t.Optional[str] = None) -> "AbstractFile":
        """Open a file"""
        entry = self.get_file_entry(fullname)
        return entry.open(file_mode)

    def read_bytes(self, fullname: str, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of a file"""
        entry = self.get_file_entry(fullname)
        return entry.read_bytes(file_mode)

    def read_text(self, fullname: str, encoding: str = "ascii", errors: str = "ignore", file_mode: str = ASCII) -> str:
        """Get the content of a file as text"""
        data = self.read_bytes(fullname, file_mode)
        return data.decode(encoding, errors)

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        # TODO: Check block range
        if fullname:
            if start is None:
                start = 0
            if end is None:
                entry = self.get_file_entry(fullname)
                end = entry.get_length() - 1
            f = self.open_file(fullname, file_mode=IMAGE)
            try:
                for block_number in range(start, end + 1):
                    data = f.read_block(block_number)
                    sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                    hex_dump(data)
            finally:
                f.close()
        elif hasattr(self, "read_block"):
            if start is None:
                start = 0
            if end is None:
                if start == 0:
                    end = self.get_size() // BLOCK_SIZE
                else:
                    end = start
            for block_number in range(start, end + 1):
                data = self.read_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                hex_dump(data)

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return []


class AbstractBlockFilesystem(AbstractFilesystem):
    """Abstract base class for block-based filesystems"""

    dev: BlockDevice

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = BlockDevice(file_or_device)
        elif isinstance(file_or_device, BlockDevice):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, "Not a valid block device")

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        return self.dev.read_block(block_number, number_of_blocks)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        self.dev.write_block(buffer, block_number, number_of_blocks)



from xferx.commons import IMAGE
from xferx.pdp7.codes import (
    fiodec_to_str,
    read_baudot_string,
    str_to_baudot,
    str_to_fiodec,
)
from xferx.pdp7.decsysfs import (  # LibraryDirectory,
    DECSysDirectoryEntry,
    DECSysFilesystem,
    ProgramDirectory,
    from_bytes_to_18bit_words,
)

# from xferx.shell import Shell


class MockFilesytem(DECSysFilesystem):

    def __init__(self, t):
        self.t = t

    def read_words_block(self, block_number):
        if block_number in self.t:
            return list(self.t[block_number])
        else:
            raise ValueError(f"Invalid block number {block_number}")

    def write_words_block(self, block_number, words) -> None:
        self.t[block_number] = list(words)


def test_baudot():
    strings = [
        "A",
        "1.",
        "HELLO",
        "FORTRN",
    ]
    for s in strings:
        words = str_to_baudot(s, length=None)
        s1, _ = read_baudot_string(words)
        assert s1 == s
    for s in strings:
        words = str_to_baudot(s, length=2)
        assert len(words) == 2
        s1, _ = read_baudot_string(words)
        assert s1 == s
    l0 = "CAB DECSYS7 COPY"
    l1 = "15 JUNE 1966"
    words = str_to_baudot(l0) + [0o777777] + str_to_baudot(l1) + [0o777777]
    s0, p = read_baudot_string(words)
    assert s0 == l0
    s1, p = read_baudot_string(words, p + 1)
    assert s1 == l1
    assert p == len(words) - 1


def test_directory_entry():
    fs = MockFilesytem({})
    directory = ProgramDirectory(fs)
    t = [1, 115084, 10252, 7, 64]
    entry = DECSysDirectoryEntry.read(directory, t, 0)
    assert entry.file_type == "SYSTEM"
    assert entry.filename == "CONTEN"
    assert entry.block_number == 7
    assert entry.starting_address == 63
    assert entry.to_words() == t

    t = [1, 76838, 132224, 9, 64]
    entry = DECSysDirectoryEntry.read(directory, t, 0)
    assert entry.file_type == "SYSTEM"
    assert entry.filename == "LABEL"
    assert entry.block_number == 9
    assert entry.starting_address == 63
    assert entry.to_words() == t

    t = [2, 10280, 8192, 146, 147, 148]
    entry = DECSysDirectoryEntry.read(directory, t, 0)
    assert entry.file_type == "WORKING"
    assert entry.filename == "TEST"
    assert entry.fortran_block_number == 146
    assert entry.assembler_block_number == 147
    assert entry.block_number == 148
    assert entry.to_words() == t


def test_program_directory():
    # fmt: off
    program_directory = [73, 1, 115084, 10252, 7, 64, 1, 76838, 132224, 9, 64, 1, 115086, 108050, 10, 64, 1, 199208, 132006, 11, 64, 1, 74160, 147456, 12, 64, 1, 74160, 149524, 13, 4887, 1, 180628, 24744, 23, 64, 1, 183336, 164774, 35, 2859, 1, 180628, 9484, 47, 5958, 1, 231076, 196768, 67, 4994, 1, 133400, 8192, 76, 577, 2, 43026, 74112, 143, 144, 145, 2, 10280, 8192, 146, 0, 0, 2, 74124, 90112, 147, 148, 149, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 157]
    # fmt: on
    fs = MockFilesytem({2: list(program_directory)})
    directory = ProgramDirectory.read(fs)
    assert len(directory.entries) == 14
    assert directory.first_free_block == 157
    fs.write_words_block(2, [0] * 256)
    directory.write()
    assert fs.read_words_block(2) == program_directory


def test_fiodec():
    # fmt: off
    data = b'\xb3\x80\xb8\xb5\xa3\xa3\xa6\x80\x96\xa6\xa9\xa3\xb4\x80\xa7\xa9\xa6\xb7\xa9\xb1\xa4\x8c\x80\x81\x9e\x96\xa9\xb9\x93\xb5\x80\x82\x9b\x80\x81\x90\x8c\x80\x82\x81\x90\x9e\xb6\xa6\xa9\xa4\xb1\x93\x80\xaf\x81\x82\xb8\x80\xb8\xb5\xa3\xa3\xa6\x9b\x80\x96\xa6\xa9\xa3\xb4\xad\x8f\x8f\x8c\x80\x83\x9e\xb5\xa5\xb4\x8f\x8f\x8c\x80\x84\x8d\x80\x81\x8e\x80\x80\x80\x80\x80'
    # fmt: on
    words = from_bytes_to_18bit_words(data, file_type=IMAGE)
    txt = fiodec_to_str(words)
    words2 = str_to_fiodec(txt)
    txt2 = fiodec_to_str(words2)
    assert words == words2
    assert txt == txt2

    # fmt: off
    block_words = [
        0o000000, 0o777742, 0o630070, 0o654343, 0o460026, 0o465143, 0o640047, 0o514667,
        0o516144, 0o140001, 0o362651, 0o712365, 0o000233, 0o000120, 0o140002, 0o012036,
        0o664651, 0o446123, 0o005701, 0o027000, 0o706543, 0o434633, 0o002646, 0o514364,
        0o551717, 0o140003, 0o366545, 0o641717, 0o140004, 0o150001, 0o160000, 0o000000,
    ] + [0] * (256-32)
    # fmt: on
    assert len(block_words) == 256
    num_words = 0x40000 - block_words[1]
    words3 = block_words[2 : 2 + num_words]
    txt3 = fiodec_to_str(words3)
    words4 = str_to_fiodec(txt3)
    assert words3 == words4
    assert words3 == words

    # fmt: off
    data2 = b'\xb3\x80\xb8\xb5\xa3\xa3\xa6\x80\x96\xa6\xa9\xa3\xb4\x80\xa7\xa9\xa6\xb7\xa9\xb1\xa4\x8c\x80\x81\x9e\x96\xa9\xb9\x93\xb5\x80\x82\x9b\x80\x81\x90\x8c\x80\x82\x81\x90\x9e\xb6\xa6\xa9\xa4\xb1\x93\x80\xaf\x81\x82\xb8\x80\xb8\xb5\xa3\xa3\xa6\x9b\x80\x96\xa6\xa9\xa3\xb4\xad\x8f\x8f\x8c\x80\x83\x9e\xb5\xa5\xb4\x8f\x8f\x8c\x80\x84\x8d\x80\x81\x8e\x80\x80\x80\x80\x80'
    # fmt: on
    words5 = from_bytes_to_18bit_words(data2, file_type=IMAGE)
    txt5 = fiodec_to_str(words5)
    words6 = str_to_fiodec(txt5)
    txt6 = fiodec_to_str(words6)
    assert words5 == words6
    assert txt5 == txt6


# DSK = "tests/dsk/unixv0.dsk"
#
#
# def test_unix0_read():
#     shell = Shell(verbose=True)
#     shell.onecmd(f"mount t: /unix0 {DSK}", batch=True)
#     fs = shell.volumes.get('T')
#     assert isinstance(fs, UNIX0Filesystem)
#     assert fs.version == 0
#
#     shell.onecmd("dir t:", batch=True)
#     shell.onecmd("dir t:/", batch=True)
#     shell.onecmd("dir t:/system/", batch=True)
#     shell.onecmd("type t:/system/password", batch=True)
#
#     x = fs.read_text("dd/data/9k")
#     assert x.startswith("|")
#
#     l = list(fs.entries_list)
#     filenames = [x.filename for x in l if not x.is_empty]
#     assert "dd" in filenames
#     assert "system" in filenames
#
#     entry = fs.get_file_entry("/test/a")
#     assert not entry.inode.is_large
#
#     entry = fs.get_file_entry("/test/b")
#     assert not entry.inode.is_large
#
#     entry = fs.get_file_entry("/test/c")
#     assert entry.inode.is_large



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import re
import sys
import typing as t
from abc import abstractmethod
from datetime import date, datetime

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, BLOCK_SIZE, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.block_12bit import BlockDevice12Bit, RXBlockDevice12Bit
from ..uic import ANY_GROUP, ANY_USER, UIC
from .os8fs import oct_dump

__all__ = [
    "TSS8File",
    "TSS8Filesystem",
]

# TSS/8 TIME-SHARING SYSTEM USER'S GUIDE, Pag 117
# https://bitsavers.org/pdf/dec/pdp8/tss8/DEC-T8-MRFB-D_UserGde_Feb70.pdf

# System Manager's Guide for PDP-8E TSS 8.24 Monitor, Pag 147
# https://bitsavers.org/pdf/dec/pdp8/tss8/TSS8_8.24_ManagersGuide.pdf

# The disk is divided into tracks. One track is defined as 4096 words.

WORDS_PER_BLOCK = 256  # One segment is defined as 256 words.
WORDS_PER_TRACK = 4096  # Track size (in words)
BLOCKS_PER_TRACK = WORDS_PER_TRACK // WORDS_PER_BLOCK  # Number of blocks per track
MONITOR_SIZE = 5 * BLOCKS_PER_TRACK  # Monitor size (in blocks)
SI_BLOCK = 0 * BLOCKS_PER_TRACK  # System Interpreter
FIP_BLOCK = 1 * BLOCKS_PER_TRACK  # FIP File Phantom
INIT_BLOCK = 2 * BLOCKS_PER_TRACK
TSS8_BLOCK = 3 * BLOCKS_PER_TRACK  # TSS/8 resident monitor (2 traks)

# MFD

MFD_UID_POS = 0
MFD_PASSWORD_POS = 1
MFD_PASSWORD_SIZE = 2  # Password size (in words)
MFD_NEXT_POS = 3
MFD_QUOTA_POS = 4
MFD_DEVICE_TIME_POS = 5
MFD_CPU_TIME_POS = 6
MFD_RETRIEVAL_POINTER_POS = 7
MFD_SIZE = 8  # in words

# UFD

UFD_FILENAME_SIZE = 3  # Filename size (in words)
UFD_FILENAME_POS = 0
UFD_NEXT_POS = 3
UFD_EXT_PROTECTION_POS = 4
UFD_FILE_SIZE_POS = 5
UFD_CREATION_DATE_POS = 6
UFD_RETRIEVAL_POINTER_POS = 7
UFD_SIZE = 8  # in words

# SAT

SAT_SIZE = 0o530  # Size of the Storage Allocation Table (in words)
SAT_END_POS = 0o7777  # Position of the end of the Storage Allocation Table in FIP
SAT_START_POS = SAT_END_POS - SAT_SIZE + 1 + 2  # Position of the Storage Allocation Table in FIP
SAT_CNT = SAT_END_POS - SAT_SIZE + 2  # Position of the available segments count

# Pag 23
# https://bitsavers.org/pdf/dec/pdp8/tss8/TSS8_8.24_ManagersGuide.pdf

QUOTA_MULTIPLIER = 25  # Quota multiplier
ENTRY_SIZE = 8  # MFD/UFD entry in words
RETRIEVAL_SIZE = 8  # Size of the retrieval block (in words)

assert ENTRY_SIZE == RETRIEVAL_SIZE == MFD_SIZE == UFD_SIZE

# Pag 79
# https://bitsavers.org/pdf/dec/pdp8/tss8/TSS8_8.24_ManagersGuide.pdf

DEFAULT_PROTECTION_CODE = 0o12  # Default protection code for new files

# File protection masks (octal):
#  1 - Other projects - read protected
#  2 - Other projects - write protected
#  4 - Same project   - read protected
# 10 - Same project   - write protected
# 20 - Owner          - write protected

# Pag 109
# https://bitsavers.org/pdf/dec/pdp8/tss8/TSS8_8.24_ManagersGuide.pdf

EXTENSIONS = [
    "",
    "ASC",  #   1 - ASCII files
    "SAV",  #   2 - Save format files
    "BIN",  #   3 - Binary files
    "BAS",  #   4 - BASIC source files
    "BAC",  #   5 - BASIC compiled files
    "FCL",  #   6 - FOCAL source files
    "TMP",  #   7 - Temporary files
    "",  #      8 - blank
    "DAT",  #   9 - Data files
    "LST",  #  10 - Listing files
    "PAL",  #  11 - PAL assembler source files
    "",  #     12 - blank
    "",  #     13 - blank
    "",  #     14 - blank
    "",  #     15 - blank
]

PARTITION_FULLNAME_RE = re.compile(r"^\[(\d+)\](.*)$")
BINARY_EXTENSIONS = {"SAV", "BIN", "BAC", "TMP", "DAT"}
TSS8_BLOCK_SIZE_BYTES = 384  # Block size (in bytes)


class PPN(UIC):
    """
    TSS/8 Project-Programmer Numbers
    """

    @classmethod
    def from_str(cls, code_str: str) -> "PPN":
        code_str = code_str.split("[")[1].split("]")[0]
        group_str, user_str = code_str.split(",")
        if group_str == "*":
            group = ANY_GROUP
        else:
            group = int(group_str, 8) & 0o77
        if user_str == "*":
            user = ANY_USER
        else:
            user = int(user_str, 8) & 0o77
        return cls(group, user)

    @classmethod
    def from_word(cls, code_int: int) -> "PPN":
        group = code_int >> 6
        user = code_int & 0o77
        return cls(group, user)

    def to_word(self) -> int:
        return (self.group << 6) + self.user


ANY_PPN = PPN.from_str("[*,*]")
DEFAULT_PPN = PPN.from_str("[0,1]")
MFD_PPN = PPN.from_str("[0,1]")


def from_12bit_words_to_ascii(words: list[int]) -> str:
    """
    Convert 12bit words to ASCII string
    """
    result = bytearray()
    for word in words:
        if word == 0:
            continue
        result.append(((word >> 6) & 0o77) + 0o40)
        result.append((word & 0o77) + 0o40)
    return result.decode("ascii", errors="replace")


def from_ascii_to_12bit_words(data: str) -> list[int]:
    """
    Convert ASCII string to 12bit words
    """
    buffer = bytearray(data.upper(), "ascii")
    if len(buffer) % 2 != 0:
        buffer.append(0x20)
    words = []
    for i in range(0, len(buffer), 2):
        chr1 = (buffer[i] - 0o40) & 0o77
        chr2 = (buffer[i + 1] - 0o40) & 0o77
        words.append(chr2 | (chr1 << 6))
    return words


def tss8_to_date(val: t.Optional[int]) -> t.Optional[date]:
    """
    Translate TSS/8 date to Python date
    """
    if val is None:
        return None
    day = val % 31 + 1
    month = (val // 31) % 12 + 1
    year = val // 372 + 1974
    try:
        return date(year, month, day)
    except:
        return None


def date_to_tss8(d: t.Optional[date]) -> int:
    """
    Convert Python date to TSS/8 date integer format
    """
    if d is None:
        return 0
    return ((d.year - 1974) * 372 + (d.month - 1) * 31 + d.day - 1) & 0o7777


def tss8_canonical_filename(fullname: str, wildcard: bool = False) -> str:
    # Split the PPN from the fullname if it exists
    try:
        if "[" in fullname:
            ppn: t.Optional[PPN] = PPN.from_str(fullname)
            fullname = fullname.split("]", 1)[1]
        else:
            ppn = None
    except Exception:
        ppn = None
    if fullname:
        fullname = fullname.upper()
        try:
            filename, extension = fullname.split(".", 1)
        except Exception:
            filename = fullname
            extension = "*" if wildcard else ""
        filename = from_12bit_words_to_ascii(from_ascii_to_12bit_words(filename)[:UFD_FILENAME_SIZE]).strip()
        # TODO
        # check extension
        fullname = f"{filename}.{extension}"
    return f"{ppn or ''}{fullname}"


def tss8_split_fullname(
    ppn: PPN,
    fullname: t.Optional[str],
    wildcard: bool = True,
) -> t.Tuple[PPN, t.Optional[str]]:
    if fullname:
        if "[" in fullname:
            try:
                ppn = PPN.from_str(fullname)
                fullname = fullname.split("]", 1)[1]
            except Exception:
                return ppn, fullname
        if fullname:
            fullname = tss8_canonical_filename(fullname, wildcard=wildcard)
    return ppn, fullname


def tss8_prepare_filename_extension(filename: str) -> t.Tuple[str, str, int]:
    """
    Prepare the filename and extension for TSS/8
    """
    filename = tss8_canonical_filename(filename, wildcard=False)
    if not filename:
        raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))
    try:
        filename, extension = filename.split(".", 1)
    except Exception:
        extension = ""
    try:
        extension_idx = EXTENSIONS.index(extension)
    except Exception:
        extension_idx = 0
        extension = ""
    return filename, extension, extension_idx


def from_12bit_words_to_bytes(words: list[int], file_mode: str = ASCII) -> bytes:
    """
    Convert 12bit words to bytes
    """
    mask = 127 if file_mode == ASCII else 255
    data = bytearray()
    while words:
        if len(words) < 2:
            break
        dw = (words.pop(0) << 12) + words.pop(0)
        data.append((dw >> 16) & mask)
        data.append((dw >> 8) & mask)
        data.append(dw & mask)
    return bytes(data)


def from_bytes_to_12bit_words(data: bytes, file_mode: str = ASCII) -> list[int]:
    """
    Convert bytes back to 12-bit words
    """
    mask = 127 if file_mode == ASCII else 255
    words = []
    i = 0
    while i + 2 < len(data):
        b1, b2, b3 = data[i] & mask, data[i + 1] & mask, data[i + 2] & mask
        dw = (b1 << 16) | (b2 << 8) | b3
        words.append((dw >> 12) & 0xFFF)
        words.append(dw & 0xFFF)
        i += 3
    return words


def format_time(seconds: int) -> str:
    """
    Format seconds into HH:MM:SS
    """
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return f"{hours:02}:{minutes:02}:{seconds:02}"


class TSS8File(AbstractFile):
    entry: "UserFileDirectoryEntry"
    file_mode: str
    closed: bool
    size: int

    def __init__(self, entry: "UserFileDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        if file_mode is not None:
            self.file_mode = file_mode
        else:
            self.file_mode = IMAGE if entry.extension.upper() in BINARY_EXTENSIONS else ASCII
        self.closed = False
        self.size = entry.length * TSS8_BLOCK_SIZE_BYTES

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Read the blocks
        for disk_block_number in blocks:
            words = self.entry.ufd.fs.read_words_block(disk_block_number)
            t = from_12bit_words_to_bytes(words, self.file_mode)
            data.extend(t)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words = from_bytes_to_12bit_words(buffer, self.file_mode)
        # Get the blocks to be written
        blocks = list(self.entry.blocks())[block_number : block_number + number_of_blocks]
        # Write the blocks
        for i, disk_block_number in enumerate(blocks):
            block_words = words[i * WORDS_PER_BLOCK : (i + 1) * WORDS_PER_BLOCK]
            block_words.extend([0] * (WORDS_PER_BLOCK - len(block_words)))
            self.entry.ufd.fs.write_words_block(disk_block_number, block_words)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class TSS8AbstractDirectoryEntry(AbstractDirectoryEntry):
    """
    Master or User File Directory Entry
    """

    position: int = 0  # Position (offset in MFD/UFD)
    next: int = 0  # Link to next entry
    retrieval_pointer: int = 0  # Retrieval pointer (offset in MFD/UFD)

    @abstractmethod
    def update_dir(self) -> None:
        """
        Write the entry in the Master/User File Directory
        """
        pass

    @abstractmethod
    def to_words(self) -> t.List[int]:
        """
        Convert the entry to a list of 12-bit words
        """
        pass


class MasterFileDirectoryEntry(TSS8AbstractDirectoryEntry):
    """
    Master File Directory Entry in the MFD

    Word
       +-----------------------+
    0  | User Identification   |  Project-Programmer Number (PPN)
       +-----------------------+
    1  | Password (2 words)    |
    2  |                       |
       +-----------------------+
    3  | Next                  |  Offset of next entry in MFD
       +-----------------------+
    4  | Quota                 |
       +-----------------------+
    5  | Device Time           |
       +-----------------------+
    6  | CPU Time              |
       +-----------------------+
    7  | Retrieval Pointer     |  Offset of the retrieval block in MFD
       +-----------------------+

    """

    mfd: "MasterFileDirectory"  # Reference to the Master File Directory
    position: int  # Position (offset in MFD)
    ppn: PPN  # Project-Programmer Number (PPN)
    password: str  # Password
    next: int  # Link to next entry
    quota: int  # Quota
    cpu_time: int  # CPU time
    device_time: int  # Device time
    retrieval_pointer: int  # Retrieval pointer (offset in MFD)

    def __init__(self, mfd: "MasterFileDirectory"):
        self.mfd = mfd

    @classmethod
    def read(cls, mfd: "MasterFileDirectory", position: int) -> "MasterFileDirectoryEntry":
        """
        Read a User File Directory from disk
        """
        self = cls(mfd)
        self.position = position
        self.ppn = PPN.from_word(mfd.words[position + MFD_UID_POS])  # Project-Programmer Number (PPN)
        password = mfd.words[position + MFD_PASSWORD_POS : position + MFD_PASSWORD_POS + MFD_PASSWORD_SIZE]
        self.password = from_12bit_words_to_ascii(password)  # Password
        self.next = mfd.words[position + MFD_NEXT_POS]  # Link to next entry
        self.quota = mfd.words[position + MFD_QUOTA_POS] * QUOTA_MULTIPLIER  # Quota
        self.device_time = mfd.words[position + MFD_DEVICE_TIME_POS]  # Device time
        self.cpu_time = mfd.words[position + MFD_CPU_TIME_POS]  # CPU time
        self.retrieval_pointer = mfd.words[position + MFD_RETRIEVAL_POINTER_POS]  # Pointer to retrieval
        return self

    @classmethod
    def create(
        cls,
        mfd: "MasterFileDirectory",
        ppn: PPN,
        position: int,
        retrieval_pointer: int,
        password: str = "",
        quota: int = 0,
    ) -> "MasterFileDirectoryEntry":
        """
        Create a new Master File Directory Entry
        """
        self = cls(mfd)
        self.position = position
        self.ppn = ppn
        self.password = password
        self.next = 0
        self.quota = quota
        self.cpu_time = 0
        self.device_time = 0
        self.retrieval_pointer = retrieval_pointer
        self.update_dir()  # Write the entry to the MFD
        return self

    def iterdir(
        self,
        pattern: t.Optional[str] = None,
        include_all: bool = False,
        wildcard: bool = False,
    ) -> t.Iterator["UserFileDirectoryEntry"]:
        ufd = UserFileDirectory.read(self)
        entry: "UserFileDirectoryEntry"
        for entry in list(ufd.entries):  # type: ignore
            if (not entry.is_dummy) and (not pattern or filename_match(entry.basename, pattern, wildcard=wildcard)):
                yield entry

    def disk_usage(self) -> int:
        """
        Get the disk usage in blocks
        """
        return sum(entry.get_length() for entry in self.iterdir())

    def update_dir(self) -> None:
        """
        Write the entry in the User File Directory
        """
        self.mfd.words[self.position : self.position + ENTRY_SIZE] = self.to_words()

    def to_words(self) -> t.List[int]:
        """
        Convert the entry to a list of 12-bit words
        """
        password = from_ascii_to_12bit_words(self.password)[:2]
        password += [0] * (MFD_PASSWORD_SIZE - len(password))
        words = [self.ppn.to_word()]  # Project-Programmer Number (PPN)
        words += password  # Password (2 words)
        words.append(self.next)  # Link to next entry
        words.append(self.quota // QUOTA_MULTIPLIER)  # Quota
        words.append(self.device_time)  # Device Time
        words.append(self.cpu_time)  # CPU Time
        words.append(self.retrieval_pointer)  # Retrieval Pointer
        return words

    def open(self, file_mode: t.Optional[str] = None) -> TSS8File:
        raise OSError(errno.EINVAL, "Invalid operation on directory")

    @property
    def is_dummy(self) -> bool:
        """
        The MFD entry (at position 0) is a dummy entry
        """
        return self.position == 0

    @property
    def fullname(self) -> str:
        return f"{self.ppn}"

    @property
    def basename(self) -> str:
        return f"{self.ppn}"

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return len(list(self.mfd.retrieval_blocks(self.retrieval_pointer)))

    def get_size(self) -> int:
        """
        Get entry size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def delete(self) -> bool:
        # Delete all entries in the UFD
        for entry in self.iterdir():
            if not entry.delete():
                raise OSError(errno.EIO, os.strerror(errno.EIO))
        # Delete the UFD
        return self.mfd.delete_ufd(self)

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EINVAL, "Invalid operation on directory")

    @property
    def fs(self) -> "TSS8Filesystem":
        """
        Get the filesystem associated with this MFD entry
        """
        return self.mfd.fs

    def __str__(self) -> str:
        return (
            f"{str(self.ppn):<11} {self.password:8} {self.next:>5}  {self.quota:>4}  "
            f"{self.device_time:>4}  {self.cpu_time:>4}  {self.retrieval_pointer:>4}"
        )


class UserFileDirectoryEntry(TSS8AbstractDirectoryEntry):
    """
    User File Directory Entry

    Word
       +-----------------------+
    0  | File Name (3 words)   |
    3  |                       |
       +-----------------------+
    4  | Next                  |  Offset of next entry in UFD
       +-----------------------+
    5  | Ext | Protection      |
       +-----------------------+
    6  | File Size             |
       +-----------------------+
    7  | Creation Date         |
       +-----------------------+
    8  | Retrieval Pointer     |  Offset of the retrieval block in UFD
       +-----------------------+
    """

    ufd: "UserFileDirectory"
    position: int = 0  # Position (offset in MFD)
    filename: str = ""
    extension: str = ""
    extension_idx: int = 0  # Extension index
    protection: int = 0  # Protection bits
    next: int = 0  # Link to next entry
    length: int = 0  # Length in blocks
    raw_creation_date: int = 0  # Creation date
    retrieval_pointer: int = 0  # Retrieval pointer (offset in MFD)

    def __init__(self, ufd: "UserFileDirectory"):
        self.ufd = ufd

    @classmethod
    def read(cls, ufd: "UserFileDirectory", position: int) -> "UserFileDirectoryEntry":
        self = cls(ufd)
        filename = ufd.words[UFD_FILENAME_POS + position : UFD_FILENAME_POS + position + UFD_FILENAME_SIZE]
        self.position = position
        self.filename = from_12bit_words_to_ascii(filename).strip()
        self.next = ufd.words[position + UFD_NEXT_POS]
        ext_protection = ufd.words[position + UFD_EXT_PROTECTION_POS]
        self.protection = ext_protection & 0o77
        self.extension_idx = (ext_protection >> 7) & 0xF
        self.extension = EXTENSIONS[self.extension_idx]
        self.length = ufd.words[position + UFD_FILE_SIZE_POS]
        self.raw_creation_date = ufd.words[position + UFD_CREATION_DATE_POS]
        self.retrieval_pointer = ufd.words[position + UFD_RETRIEVAL_POINTER_POS]
        return self

    @classmethod
    def create(
        cls,
        ufd: "UserFileDirectory",
        basename: str,
        number_of_blocks: int,
        position: int,
        retrieval_pointer: int,
        protection_code: int = DEFAULT_PROTECTION_CODE,
        creation_date: t.Optional[date] = None,
    ) -> "UserFileDirectoryEntry":
        """
        Create a new User File Directory Entry
        """
        self = cls(ufd)
        (self.filename, self.extension, self.extension_idx) = tss8_prepare_filename_extension(basename)
        self.protection = protection_code
        self.position = position
        self.next = 0
        self.length = number_of_blocks
        self.raw_creation_date = date_to_tss8(creation_date or date.today())
        self.retrieval_pointer = retrieval_pointer
        self.update_dir()  # Write the entry to the UFD
        return self

    def update_dir(self) -> None:
        """
        Write the entry in the User File Directory
        """
        self.ufd.words[self.position : self.position + ENTRY_SIZE] = self.to_words()

    def to_words(self) -> t.List[int]:
        """
        Convert the entry to a list of 12-bit words
        """
        words = from_ascii_to_12bit_words(self.filename)[:UFD_FILENAME_SIZE]
        words = words + [0] * (UFD_FILENAME_SIZE - len(words))  # Ensure 3 words for filename
        words.append(self.next)
        words.append(self.protection + ((self.extension_idx & 0xF) << 7))
        words.append(self.length)
        words.append(self.raw_creation_date)
        words.append(self.retrieval_pointer)
        return words

    def blocks(self) -> t.Iterator[int]:
        """
        Iterate over the blocks of the file
        """
        yield from self.ufd.retrieval_blocks(self.retrieval_pointer)

    @property
    def fullname(self) -> str:
        return f"{self.ufd.ppn or ''}{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def creation_date(self) -> t.Optional[date]:
        return tss8_to_date(self.raw_creation_date)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return TSS8_BLOCK_SIZE_BYTES

    def delete(self) -> bool:
        """
        Delete the file
        """
        return self.ufd.delete(self)

    def resize(self, number_of_blocks: int) -> None:
        """
        Resize the file to the specified size
        """
        self.ufd.resize(self, number_of_blocks)

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def open(self, file_mode: t.Optional[str] = None) -> TSS8File:
        """
        Open a file
        """
        return TSS8File(self, file_mode)

    @property
    def is_dummy(self) -> bool:
        """
        The UFD entry (at position 0) is a dummy entry
        """
        return self.position == 0

    def __str__(self) -> str:
        return (
            f"{str(self.ufd.ppn):<11} {self.basename:12}  {self.protection:02o}  {str(self.creation_date or ''):12}  "
            f"{self.position:>5}  {self.next:>5}  {self.retrieval_pointer:>5}  {self.length:>5}"
        )

    def __repr__(self) -> str:
        return str(self)


class AbstractFileDirectory:
    """
    Abstract (Master or User) File Directory
    """

    fs: "TSS8Filesystem"
    words: t.List[int]
    entries: t.List["TSS8AbstractDirectoryEntry"]

    def __init__(self, fs: "TSS8Filesystem"):
        self.fs = fs

    def retrieval_blocks(self, retrieval_pointer: int) -> t.Iterator[int]:
        """
        Retrieval block

        A retrieval block is a linked list of blocks that contain the retrieval segments.

        Word
             +-----------------------+
          0  | Next Retrieval Block  | (offset in MFD/UFD)
             +-----------------------+
          1  | Retrieval Segment 1   |
             /                       /
          7  | Retrieval Segment 7   |
             +-----------------------+
        """
        while retrieval_pointer != 0:
            next_retrieval = self.words[retrieval_pointer]  # Link to next retrieval block
            retrieval_segments = self.words[retrieval_pointer + 1 : retrieval_pointer + RETRIEVAL_SIZE]
            for ret_block in retrieval_segments:
                if ret_block != 0:
                    # Calculate the disk block number
                    yield ret_block - 1 + self.fs.mfd_block
            retrieval_pointer = next_retrieval

    def resize_retrieval_blocks(
        self,
        retrieval_pointer: int,
        number_of_blocks: int,
        free_dir_blocks: t.Optional[t.List[int]] = None,
    ) -> None:
        """
        Resize the retrieval blocks to the specified size.
        This will allocate or free blocks as necessary.
        """
        current_blocks = list(self.retrieval_blocks(retrieval_pointer))
        current_size = len(current_blocks)
        if number_of_blocks <= 0:
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))
        if number_of_blocks < current_size:
            self.reduce_retrieval_blocks(
                retrieval_pointer=retrieval_pointer,
                number_of_blocks=number_of_blocks,
                free_dir_blocks=free_dir_blocks,
            )
        elif number_of_blocks > current_size:
            self.extend_retrieval_blocks(
                retrieval_pointer=retrieval_pointer,
                extend_number_of_blocks=number_of_blocks - current_size,
                free_dir_blocks=free_dir_blocks,
            )

    def reduce_retrieval_blocks(
        self,
        retrieval_pointer: int,
        number_of_blocks: int,
        free_dir_blocks: t.Optional[t.List[int]] = None,
    ) -> None:
        """
        Reduce the retrieval blocks by freeing blocks.
        This will free the last `reduce_number_of_blocks` blocks.
        """
        bitmap = self.fs.read_bitmap()
        while retrieval_pointer != 0:
            next_retrieval = self.words[retrieval_pointer]  # Link to next retrieval block
            for i in range(retrieval_pointer + 1, retrieval_pointer + RETRIEVAL_SIZE):
                if self.words[i] != 0:
                    if number_of_blocks == 0:
                        bitmap.set_free(self.words[i])
                        self.words[i] = 0  # Clear the block
                    else:
                        number_of_blocks -= 1
            retrieval_pointer = next_retrieval
        # Write the bitmap back to disk
        bitmap.write()
        self.write()  # Write the updated retrieval blocks to disk

    def extend_retrieval_blocks(
        self,
        retrieval_pointer: int,
        extend_number_of_blocks: int,
        free_dir_blocks: t.Optional[t.List[int]] = None,
    ) -> None:
        """
        Extend the retrieval blocks by allocating new blocks.
        This will allocate new blocks and link them to the retrieval pointer.
        """
        # Allocate new blocks if needed
        bitmap = self.fs.read_bitmap()
        allocated_blocks = bitmap.allocate(extend_number_of_blocks)
        if free_dir_blocks is None:
            free_dir_blocks = self.get_free_file_directory_blocks()

        while allocated_blocks:
            next_retrieval = self.words[retrieval_pointer]  # Link to next retrieval block
            for i in range(retrieval_pointer + 1, retrieval_pointer + RETRIEVAL_SIZE):
                if self.words[i] == 0:
                    self.words[i] = allocated_blocks.pop(0)
                    if not allocated_blocks:
                        break
            if allocated_blocks and next_retrieval == 0:
                # If we still have blocks to allocate
                # we need to create a new retrieval block
                next_retrieval = free_dir_blocks.pop(0)
                # TODO - extend the directory if needed
                self.words[retrieval_pointer] = next_retrieval
                self.words[next_retrieval : next_retrieval + RETRIEVAL_SIZE] = [0] * RETRIEVAL_SIZE
            retrieval_pointer = next_retrieval
        bitmap.write()  # Write the bitmap back to disk
        self.write()  # Write the updated retrieval blocks to disk

    def free_retrieval_blocks(self, retrieval_pointer: int) -> None:
        """
        Free the retrieval blocks, updating the bitmap accordingly.
        """
        bitmap = self.fs.read_bitmap()
        while retrieval_pointer != 0:
            next_retrieval = self.words[retrieval_pointer]  # Link to next retrieval block
            retrieval_segments = self.words[retrieval_pointer + 1 : retrieval_pointer + RETRIEVAL_SIZE]
            for ret_block in retrieval_segments:
                if ret_block != 0:
                    # assert not bitmap.is_free(ret_block)
                    bitmap.set_free(ret_block)
            self.words[retrieval_pointer : retrieval_pointer + RETRIEVAL_SIZE] = [
                0
            ] * RETRIEVAL_SIZE  # Clear the retrieval block
            retrieval_pointer = next_retrieval
        bitmap.write()  # Write the bitmap back to disk
        self.write()  # Write the updated retrieval blocks to disk

    def get_used_file_directory_blocks(self) -> t.Set[int]:
        """
        Get the used blocks in the File Directory
        A File Directory block is group of 8 words longs.
        """
        result = set()
        for entry in self.entries:
            result.add(entry.position)
            if entry.retrieval_pointer != 0:
                retrieval_pointer = entry.retrieval_pointer
                result.add(retrieval_pointer)
                while self.words[retrieval_pointer] != 0:
                    retrieval_pointer = self.words[retrieval_pointer]
                    result.add(retrieval_pointer)
        return result

    def get_free_file_directory_blocks(self) -> t.List[int]:
        """
        Get the free blocks in the File Directory
        A File Directory block is group of 8 words longs.
        """
        total_space = set(range(0, len(self.words), ENTRY_SIZE))
        used_blocks = self.get_used_file_directory_blocks()
        return sorted(total_space - used_blocks)

    def read_file(self, retrieval_pointer: int) -> t.List[int]:
        result = []
        for block in self.retrieval_blocks(retrieval_pointer):
            result += self.fs.read_words_block(block)
        return result

    def write_file(self, retrieval_pointer: int, words: t.List[int]) -> None:
        for i, block in enumerate(self.retrieval_blocks(retrieval_pointer)):
            tmp = words[i * WORDS_PER_BLOCK : (i + 1) * WORDS_PER_BLOCK]
            self.fs.write_words_block(block, tmp)

    @abstractmethod
    def write(self) -> None:
        """
        Write the Master File Directory back to the disk
        """
        pass


class MasterFileDirectory(AbstractFileDirectory):
    """
    Master File Directory
    """

    @classmethod
    def read(cls, fs: "TSS8Filesystem") -> "MasterFileDirectory":
        """
        Read the Master File Directory from disk
        """
        self = cls(fs)
        self.words = fs.read_12bit_words_track(self.fs.mfd_block)
        self.entries = []
        position = 0
        while True:
            entry = MasterFileDirectoryEntry.read(self, position)
            self.entries.append(entry)
            position = entry.next
            if position == 0:
                break
        return self

    def write(self) -> None:
        """
        Write the Master File Directory back to the disk
        """
        self.fs.write_12bit_words_track(self.fs.mfd_block, self.words)

    def create_ufd(
        self,
        ppn: PPN,
        password: str = "",
    ) -> t.Optional["MasterFileDirectoryEntry"]:
        """
        Create a User File Directory
        """
        bitmap = self.fs.read_bitmap()
        free_dir_blocks = self.get_free_file_directory_blocks()

        # Create a new entry
        entry = MasterFileDirectoryEntry.create(
            mfd=self,
            ppn=ppn,
            position=free_dir_blocks.pop(0),
            retrieval_pointer=free_dir_blocks.pop(0),
            password=password,
        )

        # Update the previous entry
        prev_entry = self.entries[-1]
        prev_entry.next = entry.position
        prev_entry.update_dir()  # Write the previous entry to the UFD

        # Update the retrieval pointer
        block = bitmap.allocate_one()
        self.words[entry.retrieval_pointer : entry.retrieval_pointer + ENTRY_SIZE] = [0, block, 0, 0, 0, 0, 0, 0]

        # Write an empty block for the UFD
        self.fs.write_words_block(block - 1 + self.fs.mfd_block, [0] * WORDS_PER_BLOCK)

        # Write the entry to the MFD
        self.write()
        # Write the bitmap back to disk
        bitmap.write()

        # Add the entry to the list of entries
        self.entries.append(entry)
        return entry

    def delete_ufd(self, entry: "MasterFileDirectoryEntry") -> bool:
        """
        Delete an User File Directory
        """
        try:
            # Re-read the directory
            self.words = self.fs.read_12bit_words_track(self.fs.mfd_block)
            # Get the index of the entry to be deleted
            index = self.entries.index(entry)
            # Get the previous entry
            prev_entry = self.entries[index - 1]
            # Remove the entry from the list
            del self.entries[index]
            self.words[entry.position : entry.position + ENTRY_SIZE] = [0] * ENTRY_SIZE
            # Update the next pointer of the previous entry
            prev_entry.next = entry.next
            # Write the previous entry
            self.words[prev_entry.position : prev_entry.position + ENTRY_SIZE] = prev_entry.to_words()
        except ValueError:
            # If the entry is not found, it means it has already been deleted or does not exist
            return False
        # Free the retrieval blocks, updating the bitmap accordingly
        self.free_retrieval_blocks(entry.retrieval_pointer)
        self.write()
        return True


class UserFileDirectory(AbstractFileDirectory):
    """
    User File Directory
    """

    mfd_entry: "MasterFileDirectoryEntry"

    @classmethod
    def read(cls, mfd_entry: "MasterFileDirectoryEntry") -> "UserFileDirectory":
        """
        Read the User File Directory from disk
        """
        self = cls(mfd_entry.fs)
        self.mfd_entry = mfd_entry
        self.words = mfd_entry.mfd.read_file(mfd_entry.retrieval_pointer)
        self.entries = []
        # Read the User File Directory entries
        position = 0
        while True:
            entry = UserFileDirectoryEntry.read(self, position)
            self.entries.append(entry)
            position = entry.next
            if position == 0:
                break
        return self

    def write(self) -> None:
        """
        Write the User File Directory back to disk
        """
        self.mfd_entry.mfd.write_file(self.mfd_entry.retrieval_pointer, self.words)

    def delete(self, entry: "UserFileDirectoryEntry") -> bool:
        """
        Delete a file from the User File Directory
        """
        try:
            # Re-read the words from the MFD entry
            self.words = self.mfd_entry.mfd.read_file(self.mfd_entry.retrieval_pointer)
            # Get the index of the entry to be deleted
            index = self.entries.index(entry)
            # Get the previous entry
            prev_entry = self.entries[index - 1]
            # Remove the entry from the list
            del self.entries[index]
            self.words[entry.position : entry.position + ENTRY_SIZE] = [0] * ENTRY_SIZE
            # Update the next pointer of the previous entry
            prev_entry.next = entry.next
            # Write the previous entry
            self.words[prev_entry.position : prev_entry.position + ENTRY_SIZE] = prev_entry.to_words()
        except ValueError:
            # If the entry is not found, it means it has already been deleted or does not exist
            return False
        # Free the retrieval blocks, updating the bitmap accordingly
        self.free_retrieval_blocks(entry.retrieval_pointer)
        self.write()
        return True

    def resize(self, entry: "UserFileDirectoryEntry", number_of_blocks: int) -> None:
        """
        Resize the file to the specified size
        """
        if number_of_blocks < 0:
            raise OSError(errno.EINVAL, os.strerror(errno.EINVAL))
        if number_of_blocks == 0:
            self.delete(entry)
            return
        entry.length = number_of_blocks
        entry.update_dir()  # Write the entry to the UFD
        self.resize_retrieval_blocks(entry.retrieval_pointer, number_of_blocks)

    def create_file(
        self,
        basename: str,
        number_of_blocks: int,
        protection_code: int = DEFAULT_PROTECTION_CODE,
        creation_date: t.Optional[date] = None,
    ) -> "UserFileDirectoryEntry":
        """
        Create a new file in the User File Directory
        """
        free_dir_blocks = self.get_free_file_directory_blocks()
        # TODO check UFD space

        # Create a new entry
        entry = UserFileDirectoryEntry.create(
            ufd=self,
            basename=basename,
            number_of_blocks=number_of_blocks,
            position=free_dir_blocks.pop(0),
            retrieval_pointer=free_dir_blocks.pop(0),
            protection_code=protection_code,
            creation_date=creation_date,
        )

        # Update the previous entry
        prev_entry = self.entries[-1]
        prev_entry.next = entry.position
        prev_entry.update_dir()  # Write the previous entry to the UFD

        # Update the retrieval pointer
        self.words[entry.retrieval_pointer : entry.retrieval_pointer + RETRIEVAL_SIZE] = [0] * RETRIEVAL_SIZE

        # Extend the retrieval blocks to the specified size, and write the entry/bitmap
        self.extend_retrieval_blocks(entry.retrieval_pointer, number_of_blocks, free_dir_blocks=free_dir_blocks)
        return entry

    @property
    def ppn(self) -> PPN:
        """
        Get the Project-Programmer Numbers User of the User File Directory
        """
        return self.mfd_entry.ppn


class StorageAllocationTable:
    """
    Storage Allocation Table

    SAT resides in FIP at 0o7777 and extends down through 0o7777-(SAT_SIZE-1).
    Each bit position represents 1 block of file storage.
    A block is available if its sat bit has the value 0.

    Word                               Size
           +------------------------+
    0o7250 | SATBOT                 |     1
           +------------------------+
    0o7251 | SATCNT                 |     1
           +------------------------+
    0o7252 | SAT                    | 0o530
           /                        /
    0o7777 |                        |
           +------------------------+
    """

    fs: "TSS8Filesystem"
    bitmaps: t.List[int]

    def __init__(self, fs: "TSS8Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "TSS8Filesystem") -> "StorageAllocationTable":
        """
        Read the bitmap blocks
        """
        self = StorageAllocationTable(fs)
        # Read the SAT from the FIP track
        words = self.fs.read_12bit_words_track(FIP_BLOCK)
        self.bitmaps = words[SAT_START_POS : SAT_END_POS + 1]
        return self

    def write(self) -> None:
        """
        Write the bitmap blocks
        """
        # Read the SAT from the FIP track
        words = self.fs.read_12bit_words_track(FIP_BLOCK)
        words[SAT_CNT] = self.free()  # Update the available segments count
        words[SAT_START_POS : SAT_END_POS + 1] = self.bitmaps
        # Write the SAT back to the FIP track
        self.fs.write_12bit_words_track(FIP_BLOCK, words)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 12

    def is_free(self, bit_index: int) -> bool:
        """
        Check if a block is free
        """
        int_index = bit_index // 12
        bit_position = bit_index % 12
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) == 0

    def set_used(self, bit_index: int) -> None:
        """
        Mark a block as used
        """
        int_index = bit_index // 12
        bit_position = bit_index % 12
        self.bitmaps[int_index] |= 1 << bit_position

    def set_free(self, bit_index: int) -> None:
        """
        Mark a block as free
        """
        int_index = bit_index // 12
        bit_position = bit_index % 12
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def find_contiguous_blocks(self, size: int) -> int:
        """
        Find contiguous blocks, return the first block number
        """
        current_run = 0
        start_index = -1
        for i in range(self.total_bits - 1, -1, -1):
            if self.is_free(i):
                if current_run == 0:
                    start_index = i
                current_run += 1
                if current_run == size:
                    return start_index - size + 1
            else:
                current_run = 0
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate_one(self) -> int:
        """ "
        Allocate one block
        """
        for block in range(0, self.total_bits):
            if self.is_free(block):
                self.set_used(block)
                return block
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate(self, size: int, contiguous: bool = False) -> t.List[int]:
        """
        Allocate contiguous or sparse blocks
        """
        blocks = []
        if contiguous and size != 1:
            start_block = self.find_contiguous_blocks(size)
            for block in range(start_block, start_block + size):
                self.set_used(block)
                blocks.append(block)
        else:
            for block in range(0, self.total_bits):
                if self.is_free(block):
                    self.set_used(block)
                    blocks.append(block)
                if len(blocks) == size:
                    break
            if len(blocks) < size:
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        used = 0
        for block in self.bitmaps:
            used += block.bit_count()
        return used

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return self.total_bits - self.used()

    def __str__(self) -> str:
        free = self.free()
        used = self.used()
        return f"LEFT: {free:<6} USED: {used:<6}"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, StorageAllocationTable) and self.bitmaps == other.bitmaps  # type: ignore


class TSS8Filesystem(AbstractFilesystem):
    """
    TSS/8 Filesystem

    Disk storage
                               Words
    +-----------------------+
    | Monitor               |  20k
    |                       |
    +-----------------------+
    | Swapping area         |  4k * users
    |                       |
    +-----------------------+
    | File area             |
    |                       |
    +-----------------------+

    TSS/8 TIME-SHARING SYSTEM USER'S GUIDE, Pag 117
    https://bitsavers.org/pdf/dec/pdp8/tss8/DEC-T8-MRFB-D_UserGde_Feb70.pdf

    System Manager's Guide for PDP-8E TSS 8.24 Monitor, Pag 147
    https://bitsavers.org/pdf/dec/pdp8/tss8/TSS8_8.24_ManagersGuide.pdf
    """

    fs_name = "tss8"
    fs_description = "PDP-8 TSS/8"
    fs_platforms = ["pdp-8"]
    dev: BlockDevice12Bit

    users: int  # Number of users
    mfd_block: int  # Block number of the Master File Directory
    ppn: PPN = DEFAULT_PPN

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = RXBlockDevice12Bit(file_or_device)
        elif isinstance(file_or_device, BlockDevice12Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        **kwargs: t.Union[bool, str],
    ) -> "TSS8Filesystem":
        """
        Mount the TSS/8 filesystem from a file
        """
        self = cls(file_or_dev)
        self.users, self.mfd_block = self.guess_users()
        # self.users = 20  # Default number of users
        # self.mfd_block = MONITOR_SIZE + BLOCKS_PER_TRACK * self.users
        return self

    def guess_users(self) -> t.Tuple[int, int]:
        for users in range(8, 32):
            block_number = MONITOR_SIZE + BLOCKS_PER_TRACK * users
            # block_number = (MONITOR_SIZE + 4 * users) * 4
            words = self.read_words_block(block_number)
            # The first 8-word block of the UFD is a dummy block.
            # It contains all zeros except for a pointer to the next block
            if words[UFD_NEXT_POS] != 0o10:  # Link to next block
                continue
            if words[ENTRY_SIZE + UFD_EXT_PROTECTION_POS] & 0o7700:  # Protection bits
                continue
            if words[ENTRY_SIZE + UFD_RETRIEVAL_POINTER_POS] != 0o20:  # Pointer to retrieval
                continue
            if words[ENTRY_SIZE + ENTRY_SIZE] != 0:  # PPN
                continue
            return users, block_number
        raise OSError(errno.EIO, os.strerror(errno.EIO), "No valid MFD found")

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as 256 12bit words
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 12bit words as a block
        """
        self.dev.write_words_block(block_number, words)

    def read_12bit_words_track(self, first_block_number: int) -> t.List[int]:
        """
        Read a track of 12-bit words
        """
        words = []
        for i in range(0, BLOCKS_PER_TRACK):
            words += self.read_words_block(first_block_number + i)
        return words

    def write_12bit_words_track(self, first_block_number: int, words: t.List[int]) -> None:
        """
        Write a track of 12-bit words
        """
        for i in range(0, BLOCKS_PER_TRACK):
            tmp = words[i * WORDS_PER_BLOCK : (i + 1) * WORDS_PER_BLOCK]
            self.write_words_block(first_block_number + i, tmp)

    def read_mfd_entries(
        self,
        ppn: PPN = ANY_PPN,
    ) -> t.Iterator["MasterFileDirectoryEntry"]:
        """Read Master File Directory entries"""
        mfd = MasterFileDirectory.read(self)
        entry: "MasterFileDirectoryEntry"
        for entry in mfd.entries:  # type: ignore
            if ppn.match(entry.ppn) and not entry.is_dummy:  # Filter by PPN
                yield entry

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
        ppn: t.Optional[PPN] = None,
    ) -> t.Iterator["UserFileDirectoryEntry"]:
        if ppn is None:
            ppn = self.ppn
        ppn, filename_pattern = tss8_split_fullname(fullname=pattern, wildcard=wildcard, ppn=ppn)

        if pattern and not filename_pattern and not expand:
            # If expand is False, check if the pattern is a PPN
            try:
                ppn = PPN.from_str(pattern)
                mfd = MasterFileDirectory.read(self)
                mfd_entry: "MasterFileDirectoryEntry"
                for mfd_entry in mfd.entries:  # type: ignore
                    if ppn.match(mfd_entry.ppn) and not mfd_entry.is_dummy:  # Filter by PPN
                        yield mfd_entry  # type: ignore
                return
            except Exception:
                pass

        for mfd_entry in self.read_mfd_entries(ppn=ppn):
            if ppn != MFD_PPN or include_all:
                yield from mfd_entry.iterdir(pattern=filename_pattern, include_all=include_all, wildcard=wildcard)

    @property
    def entries_list(self) -> t.Iterator["UserFileDirectoryEntry"]:
        for mfd_entry in self.read_mfd_entries(ppn=self.ppn):
            yield from mfd_entry.iterdir()

    def get_file_entry(self, fullname: str) -> "UserFileDirectoryEntry":
        """
        Get the directory entry for a file
        """
        fullname = tss8_canonical_filename(fullname)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        ppn, basename = tss8_split_fullname(fullname=fullname, wildcard=False, ppn=self.ppn)
        try:
            return next(self.filter_entries_list(basename, wildcard=False, ppn=ppn, expand=False))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def read_bitmap(self) -> StorageAllocationTable:
        """
        Read the bitmap
        """
        return StorageAllocationTable.read(self)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        number_of_blocks = int(math.ceil(len(content) / TSS8_BLOCK_SIZE_BYTES))
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        if entry is not None:
            content = content + (b"\0" * TSS8_BLOCK_SIZE_BYTES)
            f = entry.open(file_mode)
            try:
                f.write_block(content, block_number=0, number_of_blocks=entry.length)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[UserFileDirectoryEntry]:
        try:
            entry = self.get_file_entry(fullname)
            entry.resize(number_of_blocks)
            return entry
        except FileNotFoundError:
            pass
        ppn, basename = tss8_split_fullname(fullname=fullname, wildcard=False, ppn=self.ppn)
        # Get the User File Directory entry for the PPN
        try:
            mfd_entry = next(self.read_mfd_entries(ppn=ppn))
        except Exception:
            raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), str(ppn))
        ufd = UserFileDirectory.read(mfd_entry)
        # Create a new file entry in the User File Directory
        return ufd.create_file(
            basename=basename,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
        )

    def create_directory(
        self,
        fullname: str,
        options: t.Dict[str, t.Union[bool, str]],
    ) -> t.Optional["MasterFileDirectoryEntry"]:
        """
        Create a User File Directory
        """
        # Check if the PPN is valid
        try:
            ppn = PPN.from_str(fullname)
        except Exception:
            raise OSError(errno.EINVAL, "Invalid PPN")

        # Check if the PPN already exists
        if list(self.read_mfd_entries(ppn=ppn)):
            raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST))

        # Create the User File Directory
        mfd = MasterFileDirectory.read(self)
        return mfd.create_ufd(ppn=ppn)

    def chdir(self, fullname: str) -> bool:
        """
        Change the current Project-Programmer Number (PPN)
        """
        try:
            self.ppn = PPN.from_str(fullname)
            return True
        except Exception:
            return False

    def get_pwd(self) -> str:
        """
        Get the current Project-Programmer Number (PPN)
        """
        return str(self.ppn)

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        ppn, _ = tss8_split_fullname(fullname=pattern, wildcard=True, ppn=self.ppn)
        if options.get("uic") or ppn == MFD_PPN:
            if not options.get("brief"):
                dt = datetime.now().strftime('%d-%b-%y  %H:%M:%S').upper()
                sys.stdout.write(f"SYSTEM ACCOUNT    {dt}\n\n")
                sys.stdout.write(" PASSWORD    CPU        DEV     DISK  QUOTA\n\n")
            # Listing of all PPN
            for mfd in self.read_mfd_entries():
                if options.get("brief"):
                    sys.stdout.write(f"{mfd.ppn}\n")
                else:
                    cpu_time = format_time(int(mfd.cpu_time * 6.4))
                    device_time = format_time(int(mfd.device_time * 51.2))
                    disk_usage = mfd.disk_usage()
                    sys.stdout.write(
                        f"{mfd.ppn.to_word():>4o} {mfd.password:4}  {cpu_time}  {device_time} {disk_usage:>5}  {mfd.quota:>5}\n"
                    )
        else:
            blocks = 0
            if not options.get("brief"):
                dt = date.today().strftime('%d-%b-%y').upper()
                sys.stdout.write(f"DISK FILES FOR USER {ppn.group:2o},{ppn.user:2o} ON  {dt:>9}\n\n")
                sys.stdout.write("NAME      SIZE  PROT    DATE\n")
            for entry in self.filter_entries_list(pattern):
                if options.get("brief"):
                    sys.stdout.write(f"{entry.basename}\n")
                else:
                    try:
                        dt = entry.creation_date.strftime('%d-%b-%y').upper()  # type: ignore
                    except Exception:
                        dt = ""
                    blocks += entry.length
                    sys.stdout.write(
                        f"{entry.filename:<6}.{entry.extension:<3} {entry.length:>3}   {entry.protection:2o}  {dt:>9}\n"
                    )
            if not options.get("brief"):
                sys.stdout.write(f"\nTOTAL DISK SEGMENTS:  {blocks:<6}\n")
                # sys.stdout.write(f"\nTOTAL DISK SEGMENTS:  {blocks:<6} QUOTA: 1575\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            # Display the bitmap
            bitmap = self.read_bitmap()
            for i in range(0, bitmap.total_bits):
                sys.stdout.write(f"{i:>4} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
                if i % 16 == 15:
                    sys.stdout.write("\n")
            sys.stdout.write(f"\n{bitmap}\n")
        elif arg:
            sys.stdout.write("PPN         Basename     Prt  Creation        Pos   Next    Ret Length\n")
            sys.stdout.write("---         --------     ---  --------        ---   ----    --- ------\n")
            ppn, filename_pattern = tss8_split_fullname(fullname=arg, wildcard=True, ppn=self.ppn)
            if filename_pattern and "*" not in filename_pattern:
                entry = self.get_file_entry(arg)
                # sys.stdout.write(f"{entry}\n")
                sys.stdout.write(f"{entry} {list(entry.blocks())}\n")
            else:
                for entry in self.filter_entries_list(arg, include_all=False):
                    sys.stdout.write(f"{entry}\n")
        else:
            sys.stdout.write(f"Number of users:          {self.users}\n\n")
            sys.stdout.write("PPN         Password  Next Quota  Dev   CPU  Retrieval\n")
            sys.stdout.write("                                  Time  Time Pointer\n")
            sys.stdout.write("---------   --------  ---- -----  ----  ---- ---------\n")
            for mfd in self.read_mfd_entries():
                sys.stdout.write(f"{mfd}\n")

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            entry = self.get_file_entry(fullname)
            if start is None:
                start = 0
            if end is None or end > entry.get_length() - 1:
                end = entry.get_length() - 1
            blocks = list(entry.blocks())[start : end + 1]
            for i, block_number in enumerate(blocks):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {i:08}\n")
                oct_dump(words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = self.get_size() // BLOCK_SIZE - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "TSS8Filesystem":
        """
        Create an empty TSS/8 filesystem
        """
        self = cls(file_or_dev)
        self.users = 20  # Default number of users
        self.mfd_block = MONITOR_SIZE + BLOCKS_PER_TRACK * self.users

        bitmap = self.read_bitmap()
        # Mark the MFD blocks as used
        for i in range(0, BLOCKS_PER_TRACK):
            bitmap.set_used(i)
        # Mark the last block of the bitmap as used
        for i in range((self.get_size() // BLOCK_SIZE) - self.mfd_block, bitmap.total_bits):
            bitmap.set_used(i)
        bitmap.write()

        # Create the Master File Directory
        mfd = MasterFileDirectory.read(self)
        mfd.words[UFD_NEXT_POS] = 0o10  # Link to next block
        mfd.words[UFD_NEXT_POS + 8] = 0o10 + 8  # Link to next block
        mfd.create_ufd(ppn=PPN(0, 1), password="SYSTEM")
        mfd.create_ufd(ppn=PPN(0, 2), password="LIBRARY")
        mfd.create_ufd(ppn=PPN(0, 3), password="OPERATOR")
        return self

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def close(self) -> None:
        self.dev.close()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import fnmatch
import io
import math
import os
import re
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, DECTAPE, IMAGE, READ_FILE_FULL
from ..device.abstract import AbstractDevice
from ..device.block_12bit import BlockDevice12Bit, RXBlockDevice12Bit

__all__ = [
    "DMSFile",
    "DMSDirectoryEntry",
    "DMSFilesystem",
]

# PDP-8 4k Disk Monitor System
# https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf

# PDP-8 Disc System Builder
# https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-d8-sba/dec-d8-sbab-d.pdf


BLOCK_SIZE_WORDS = 129  # Block size (in words)
BYTES_PER_WORD = 2  # Each word is encoded in 2 bytes
DATA_BLOCK_SIZE_WORDS = BLOCK_SIZE_WORDS - 1  # The last word of the block is the link to the next block

DN_ENTRY_SIZE = 5  # DN entry size (words)
DN_ENTRIES = 25  # Number of directory entries
DN_START = 0o177  # DN start block number
SAM_START = DN_START + 1  # SAM start block number

DN_BLOCKS = [DN_START, DN_START + 2, DN_START + 3]
SAM_BLOCK = [SAM_START]  # SAM blocks
SCRATCH_BLOCKS = [0o373, 0o374, 0o375]  # [251, 252, 253, 254, 255]

DECTAPE_SCRATCH_BLOCKS = [0o5, 0o6, 0o7]  # DECTAPE scratch blocks
DECTAPE_DN_BLOCKS = [DN_START, 0o201, 0o207]  # DECTAPE DN blocks
DECTAPE_SAM_BLOCKS = [SAM_START, 0o203, 0o204, 0o205, 0o206, 0o207]  # DECTAPE SAM blocks

EMPTY_FILE_NUMBER = 0  # Empty file number
RESERVED_FILE_NUMBER = 1  # Reserved for monitor, DN, SAM, and scratch blocks
MAX_FILE_NUMBER = 0o77  # Max file number
MONITOR_FILENAME = "EX C"  # Monitor file name
INVALID_FILENAMES = set(["CALL", "SAVE"])  # Invalid filenames

FILE_TYPE_ASCII = 0o0  # ASCII file (6-bit ASCII)
FILE_TYPE_BIN = 0o1  # Binary file
FILE_TYPE_FTC_BIN = 0o2  # Fortran binary file
FILE_TYPE_SYS_USER = 0o3  # Saved file (SYS or USER)

EXT_SYS = "SYS"  # System file extension
EXT_USER = "USER"  # User file extension
EXT_ASCII = "ASCII"  # ASCII file extension
EXT_BINARY = "BINARY"  # Binary file extension
EXT_FTC_BIN = "FTC BIN"  # Fortran binary file extension (yes, with a space)

RE_FILENAME = re.compile(r"([^!;:]+)\s*(?:\:(\s*\d+))?(?:;(\s*\d+))?")


class DMSFilename:
    """
    Parse the canonical PDP-8 4k Disk Monitor System filename
    """

    filename: str = ""
    program_type: int = 0
    system_program: bool = False
    core_addr: int = 0o200
    entry_point: int = 0

    def __init__(self, fullname: str, file_type: t.Optional[str] = None, wildcard: bool = False):
        """
        Filenames are limited to four characters and can be composed of
        any combination of alphanumeric characters or special characters.
        Extensions are automatically appended and depend on the file type.

        The extensions are:
        - SYS (n) Saved system program file in core bank n
        - USER (n) Saved user program file in core bank n
        - ASCII Source language program file
        - BINARY Binary program file (output from PAL-D Assembler)
        - FTC BIN Interpretive binary file (output from FORTRAN Compiler)

        The filename can be followed by a core address and an entry point:
        filename.extension[:core_addr][;entry_point]
        """
        fullname = fullname.upper()
        match = RE_FILENAME.match(fullname)
        if match:
            fullname = match.group(1)
            if match.group(2):
                self.core_addr = int(match.group(2), 8)
            if match.group(3):
                self.entry_point = int(match.group(3), 8)
        try:
            filename, extension = fullname.rsplit(".", 1)
            extension = extension.upper()
        except:
            filename = fullname
            extension = None
        self.filename = sixbit_word12_to_asc(asc_to_sixbit_word12(filename[0:2]))
        self.filename += sixbit_word12_to_asc(asc_to_sixbit_word12(filename[2:4]))
        # A file name cannot be one of the following: "CALL", "SAVE"
        if self.filename in INVALID_FILENAMES:
            raise OSError(errno.EINVAL, "Invalid filename")
        if file_type == ASCII:
            self.program_type = FILE_TYPE_ASCII
            self.system_program = False
        elif wildcard and extension == "*":
            self.program_type = -1  # Wildcard
        elif extension == EXT_ASCII:
            self.program_type = FILE_TYPE_ASCII
            self.system_program = False
        elif extension == EXT_BINARY:
            self.program_type = FILE_TYPE_BIN
            self.system_program = False
        elif extension == EXT_FTC_BIN:
            self.program_type = FILE_TYPE_FTC_BIN
            self.system_program = False
        elif extension == EXT_SYS:
            self.program_type = FILE_TYPE_SYS_USER
            self.system_program = True
        elif extension == EXT_USER:
            self.program_type = FILE_TYPE_SYS_USER
            self.system_program = False
        else:
            raise OSError(errno.EINVAL, "Invalid file extension")

    def match(self, entry: "DMSDirectoryEntry") -> bool:
        """
        Match the filename with a directory entry
        """
        result = fnmatch.fnmatch(entry.filename.strip(), self.filename.strip())
        if self.program_type != -1:
            result &= self.program_type == entry.program_type and self.system_program == entry.system_program
        return result


def from_12bit_words_to_bytes(words: list[int], file_mode: str = ASCII) -> bytes:
    """
    Convert 12bit words to bytes

    https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf Pag 105
    """
    result = bytearray()

    if file_mode == ASCII:
        esc = False
        eof = False
        for word in words:
            if word == 0:
                continue
            h = word & 0o77
            l = (word >> 6) & 0o77
            for ch in (l, h):
                if esc:
                    if ch == 0o77:  # ? - Question mark
                        result.append(0o77)
                    elif ch == 0x09:  # I - Tab
                        result.append(0x09)
                    elif ch == 0x0A:  # J - LF (Line Feed)
                        result.append(0x0A)
                    elif ch == 0x0C:  # FF (Form Feed)
                        eof = True
                        break  # end of file
                    elif ch == 0x0D:  # M - CR (Carriage Return)
                        pass
                    # else:
                    #     result.append(32)
                    #     result += f"[{ch:02x}]".encode("ascii")
                    #     result.append(32)
                    esc = False
                elif ch == 0o77:
                    esc = True
                else:
                    if ch < 32:
                        ch += 64
                    result.append(ch)
            if eof:
                break
    else:
        for i in range(0, len(words), 2):
            chr1 = words[i]
            try:
                chr2 = words[i + 1]
            except IndexError:
                chr2 = 0
            chr3 = ((chr2 >> 8) & 0o17) | ((chr1 >> 4) & 0o360)
            result.append(chr1 & 0xFF)
            result.append(chr2 & 0xFF)
            result.append(chr3 & 0xFF)

    return bytes(result)


def from_bytes_to_12bit_words(byte_data: bytes, file_mode: str = "ASCII") -> t.List[int]:
    """
    Convert bytes to 12-bit words.
    """
    words = []

    if file_mode == ASCII:
        buffer = []
        for byte in byte_data:
            byte = byte & 0o177
            if byte == 0x0A:  # LF (Line Feed) => LF + CR
                buffer.append(0o77)
                buffer.append(0x0D)
                buffer.append(0o77)
                buffer.append(0x0A)
            elif byte == 0x0D:  # CR (Carriage Return)
                pass
            elif byte in (0o77, 0x09, 0x0C):
                buffer.append(0o77)
                buffer.append(byte)
            else:
                if byte > 64:
                    byte -= 64
                buffer.append(byte)
        for i in range(0, len(buffer), 2):
            l = buffer[i]
            try:
                h = buffer[i + 1]
            except IndexError:
                h = 0
            word = (h & 0o77) | ((l & 0o77) << 6)
            words.append(word)

    else:
        for i in range(0, len(byte_data), 3):
            chr1 = byte_data[i] & 0xFF
            try:
                chr2 = byte_data[i + 1] & 0xFF
            except IndexError:
                chr2 = 0
            try:
                chr3 = byte_data[i + 2] & 0xFF
            except IndexError:
                chr3 = 0
            words.append(chr1 | ((chr3 & 0o360) << 4))
            words.append(chr2 | ((chr3 & 0o17) << 8))
    return words


def sixbit_word12_to_asc(val: int) -> str:
    """
    Convert six bit ASCII 12 bit word to 2 chars of ASCII
    """
    h = val & 0o77
    l = (val >> 6) & 0o77
    return chr(l + 32) + chr(h + 32)


def asc_to_sixbit_word12(val: str) -> int:
    """
    Convert 2 chars of ASCII back to six bit ASCII 12 bit word
    """
    if len(val) >= 1:
        l = ord(val[0].upper()) - 32
    else:
        l = 0
    if len(val) >= 2:
        h = ord(val[1].upper()) - 32
    else:
        h = 0
    return (l << 6) | h


def oct_dump(words: t.List[int], words_per_line: int = 8) -> None:
    """
    Display contents in octal
    """
    for i in range(0, len(words), words_per_line):
        line = words[i : i + words_per_line]
        ascii_str = "".join([chr(x) if 32 <= x <= 126 else "." for x in from_12bit_words_to_bytes(line)])
        oct_str = " ".join([f"{x:04o}" for x in line])
        sys.stdout.write(f"{i:08o}   {oct_str.ljust(5 * words_per_line)}  {ascii_str}\n")


class DMSFile(AbstractFile):
    entry: "DMSDirectoryEntry"
    file_mode: str
    closed: bool

    def __init__(self, entry: "DMSDirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        if file_mode is not None:
            self.file_mode = file_mode
        else:
            self.file_mode = ASCII if entry.extension.upper() == EXT_ASCII else IMAGE
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        length = self.entry.get_length()
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = length
        if self.closed or block_number < 0 or number_of_blocks < 0 or block_number + number_of_blocks > length:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.entry.get_blocks())[block_number : block_number + number_of_blocks]
        for disk_block_number in blocks:
            words = self.entry.dn.fs.read_words_block(disk_block_number)
            # Skip the last word of the block (the link to the next block)
            t = from_12bit_words_to_bytes(words[:-1], self.file_mode)
            data.extend(t)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words = from_bytes_to_12bit_words(buffer, self.file_mode)
        # Get the blocks to be written
        blocks = list(self.entry.get_blocks())[block_number : block_number + number_of_blocks]
        # Write the blocks
        for i, disk_block_number in enumerate(blocks):
            block_words = words[i * DATA_BLOCK_SIZE_WORDS : (i + 1) * DATA_BLOCK_SIZE_WORDS]
            block_words.extend([0] * (BLOCK_SIZE_WORDS - len(block_words)))
            # The last word of the block is the link to the next block
            block_words[-1] = blocks[i + 1] if i + 1 < len(blocks) else 0
            self.entry.dn.fs.write_words_block(disk_block_number, block_words)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class StorageAllocationMapBlock:
    """
    Storage Allocation Map (SAM) Block

    SAM blocks contain a record of which files are occupying
    which blocks on the device.

    https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf Pag 103

               6 bits            6 bits
        +-----------------------------------+
      1 | Block 128       |         Block 0 |
        | ...             |             ... |
        | Block 255       |       Block 127 |
        +-----------------------------------+
        | Link to next SAM block            |
        +-----------------------------------+

    The value of each half-work is the internal file number of the file
    occupying that block.

    Special values:
    0 - Block is free
    1 - Monitor, DN, SAM and scratch blocks
    4 - Loader blocks
    5 - Command decoder blocks
    """

    fs: "DMSFilesystem"
    # Block number
    block_number: int
    # Block sequence number
    block_seq_nr: int = 0
    # Link to next SAM block
    next_sam_block_number: int = 0
    # Storage Allocation Map
    sam: t.List[int] = []  # Block number => file number

    def __init__(self, fs: "DMSFilesystem"):
        self.fs = fs
        self.sam = [0] * 256

    @classmethod
    def read(cls, fs: "DMSFilesystem", block_number: int, block_seq_nr: int) -> "StorageAllocationMapBlock":
        """
        Read a Storage Allocation Map (SAM) Block
        """
        self = cls(fs)
        self.block_number = block_number
        self.block_seq_nr = block_seq_nr
        words = self.fs.read_words_block(block_number)
        for i in range(0, 128):
            self.sam[i] = words[i] & 0o77
            self.sam[i + 128] = words[i] >> 6
        self.next_sam_block_number = words[128]
        return self

    @classmethod
    def new(cls, fs: "DMSFilesystem", block_number: int, block_seq_nr: int) -> "StorageAllocationMapBlock":
        """
        Create a new Storage Allocation Map (SAM) Block
        """
        self = cls(fs)
        self.block_number = block_number
        self.block_seq_nr = block_seq_nr
        self.sam = [0] * 256
        self.next_sam_block_number = 0
        return self

    def write(self) -> None:
        """
        Write the Storage Allocation Map (SAM) Block
        """
        words = []
        assert len(self.sam) == 256
        for i in range(0, 128):
            words.append((self.sam[i] & 0o77) | ((self.sam[i + 128] & 0o77) << 6))
        words.append(self.next_sam_block_number)
        self.fs.write_words_block(self.block_number, words)

    def set_block(self, block_number: int, file_number: int) -> None:
        """
        Allocate a block to a file
        """
        self.sam[block_number - self.block_seq_nr * 256] = file_number

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return self.sam.count(EMPTY_FILE_NUMBER)

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write("\n*SAM Block\n")
        buf.write(f"Block number:          {self.block_number:>5}\n")
        buf.write(f"Next SAM block:        {self.next_sam_block_number:>5}\n")
        offset = self.block_seq_nr * 256
        for i, b in enumerate(self.sam):
            if i % 8 == 0:
                buf.write("\n")
            buf.write(f"{i+offset:>4}:{b:>2}  ")
        return buf.getvalue()


class StorageAllocationMap:
    """
    Storage Allocation Map (SAM)
    """

    fs: "DMSFilesystem"
    files_blocks: t.Dict[int, t.List[int]]  # file number => blocks
    sam_blocks: t.List["StorageAllocationMapBlock"]

    def __init__(self, fs: "DMSFilesystem"):
        self.fs = fs
        self.files_blocks = {}

    @classmethod
    def read(cls, fs: "DMSFilesystem") -> "StorageAllocationMap":
        self = StorageAllocationMap(fs)
        self.sam_blocks = []
        for sam_nr, sam in enumerate(self.read_storage_allocation_map_blocks()):
            for block_number, file_number in enumerate(sam.sam, start=sam_nr * 256):
                if file_number:
                    if file_number not in self.files_blocks:
                        self.files_blocks[file_number] = []
                    self.files_blocks[file_number].append(block_number)
            self.sam_blocks.append(sam)
        return self

    @classmethod
    def new(cls, fs: "DMSFilesystem", sam_blocks: t.List[int]) -> "StorageAllocationMap":
        self = StorageAllocationMap(fs)
        self.sam_blocks = []
        self.files_blocks = {}
        # Create SAM blocks
        for i, block_number in enumerate(sam_blocks):
            sam_block = StorageAllocationMapBlock.new(self.fs, block_number, i)
            if i < len(sam_blocks) - 1:
                sam_block.next_sam_block_number = sam_blocks[i + 1]
            self.sam_blocks.append(sam_block)
        return self

    def write(self) -> None:
        """
        Write the Storage Allocation Map (SAM) Blocks
        """
        for sam in self.sam_blocks:
            sam.write()

    def read_storage_allocation_map_blocks(self) -> t.Iterator["StorageAllocationMapBlock"]:
        """
        Read the Storage Allocation Map (SAM) Blocks
        """
        next_block_number = self.fs.first_sam_block_number
        block_seq_nr = 0
        while next_block_number != 0:
            sam = StorageAllocationMapBlock.read(self.fs, next_block_number, block_seq_nr)
            next_block_number = sam.next_sam_block_number
            block_seq_nr += 1
            yield sam

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        free_blocks = 0
        for sam in self.sam_blocks:
            free_blocks += sam.free()
        return free_blocks

    def set_block(self, block_number: int, file_number: int) -> None:
        """
        Allocate a block to a file
        """
        for sam_nr, sam in enumerate(self.sam_blocks):
            if block_number >= sam_nr * 256 and block_number < (sam_nr + 1) * 256:
                sam.set_block(block_number, file_number)
                if file_number not in self.files_blocks:
                    self.files_blocks[file_number] = []
                self.files_blocks[file_number].append(block_number)
                break

    def allocate_space(
        self,
        fullname: str,
        length: int,  # length in blocks
    ) -> int:
        """
        Allocate space for a file, return the new file number
        """
        if length > self.free():
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
        # Find a free file number
        used_file_numbers = set(self.files_blocks.keys())
        new_file_number = None
        for i in range(2, MAX_FILE_NUMBER + 1):  # Skip reserved file numbers
            if i not in used_file_numbers:
                new_file_number = i
                break
        if new_file_number is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
        # Allocate space
        blocks = []
        for sam_nr, sam in enumerate(self.sam_blocks):
            for block_number, file_number in enumerate(sam.sam, start=sam_nr * 256):
                if file_number == 0:
                    blocks.append(block_number)
                    sam.set_block(block_number, new_file_number)
                    length -= 1
                    if length == 0:
                        break
            if length == 0:
                break
        self.files_blocks[new_file_number] = blocks
        return new_file_number

    def free_space(self, file_number: int) -> None:
        """
        Free block allocated to a file
        """
        for sam_nr, sam in enumerate(self.sam_blocks):
            for block_number, fn in enumerate(sam.sam, start=sam_nr * 256):
                if fn == file_number:
                    sam.set_block(block_number, EMPTY_FILE_NUMBER)
        if file_number in self.files_blocks:
            del self.files_blocks[file_number]


class DMSDirectoryEntry(AbstractDirectoryEntry):
    """
    Directory Name Entry

    https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf Pag 102

        +-----------------------------------+
      0 | File name char 1-2                |
      1 | File name char 3-4                |
      2 | Low core addr / -1 not contiguous |
      3 | Entry point                       |
      4 | Flags                             |
        +-----------------------------------+

    The flags are as follows:
         0 1 2 3 4 5 6 7 8 9 10 11

         0-1  - Program type
         2-4  - Extended memory bits
         5    - System program
         6-11 - File number

    """

    dn: "DirectorNameBlock"
    low_core_addr: int = 0  # Core address low bits
    high_core_addr: int = 0  # Core bank (core address high bits)
    entry_point: int  # Entry point
    filename: str = ""  # Filename (4 chars)
    file_number: int = 0  # 0-63 (6 bits), 0 = Empty
    system_program: bool = False  # System/User program
    program_type: int = 0  # FILE_TYPE_ASCII, FILE_TYPE_BIN, FILE_TYPE_FTC_BIN, FILE_TYPE_SYS_USER

    def __init__(self, dn: "DirectorNameBlock", sam: "StorageAllocationMap"):
        self.dn = dn
        self.sam = sam

    @classmethod
    def read(
        cls,
        dn: "DirectorNameBlock",
        sam: "StorageAllocationMap",
        words: t.List[int],
        position: int,
    ) -> "DMSDirectoryEntry":
        self = cls(dn, sam)
        n1 = words[0 + position]  # Filename char 1-2
        n2 = words[1 + position]  # Filename char 3-4
        self.low_core_addr = words[2 + position]  # Low core addr / -1 not contiguous
        self.entry_point = words[3 + position]  # Entry point
        flags = words[4 + position]  # Flags
        self.filename = sixbit_word12_to_asc(n1) + sixbit_word12_to_asc(n2)
        self.program_type = flags >> 10
        self.high_core_addr = (flags >> 7) & 0o7
        self.file_number = flags & 0o77
        self.system_program = bool(flags >> 6 & 1)
        return self

    def to_words(self) -> t.List[int]:
        """
        Write the directory entry
        """
        flags = self.program_type << 10 | self.high_core_addr << 7 | self.system_program << 6 | self.file_number
        return [
            asc_to_sixbit_word12(self.filename[0:2]),
            asc_to_sixbit_word12(self.filename[2:4]),
            self.low_core_addr & 0o7777,
            self.entry_point & 0o7777,
            flags & 0o7777,
        ]

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        if file_mode is None:
            if self.program_type == FILE_TYPE_ASCII:
                file_mode = ASCII
            else:
                file_mode = IMAGE
        # Always read the file as IMAGE
        f = self.open(IMAGE)
        try:
            data = f.read_block(0, READ_FILE_FULL)
            if file_mode == ASCII:
                words = from_bytes_to_12bit_words(data, file_mode=IMAGE)
                return from_12bit_words_to_bytes(words, file_mode=ASCII)
            else:
                return data
        finally:
            f.close()

    @property
    def is_empty(self) -> bool:
        return self.file_number == EMPTY_FILE_NUMBER

    @property
    def extension(self) -> str:
        if self.program_type == FILE_TYPE_ASCII:  # ASCII file (6-bit ASCII)
            return EXT_ASCII
        elif self.program_type == FILE_TYPE_BIN:  # Binary file
            return EXT_BINARY
        elif self.program_type == FILE_TYPE_FTC_BIN:  # Fortran binary file
            return EXT_FTC_BIN
        else:  # Saved file (SYS or USER)
            return EXT_SYS if self.system_program else EXT_USER

    @property
    def fullname(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return self.fullname

    def get_blocks(self) -> t.List[int]:
        return self.sam.files_blocks.get(self.file_number, [])

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return len(self.get_blocks())

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        # The last word of the block is the link to the next block
        return (BLOCK_SIZE_WORDS - 1) * BYTES_PER_WORD

    @property
    def creation_date(self) -> t.Optional[date]:
        return None

    def delete(self) -> bool:
        """
        Delete the file
        """
        if self.file_number in (EMPTY_FILE_NUMBER, RESERVED_FILE_NUMBER):
            return False
        else:
            self.sam.free_space(self.file_number)
            self.sam.write()
            self.low_core_addr = 0
            self.entry_point = 0
            self.filename = ""
            self.file_number = EMPTY_FILE_NUMBER
            self.system_program = False
            self.high_core_addr = 0
            self.program_type = 0
            self.dn.write()
            return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.dn.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> DMSFile:
        """
        Open a file
        """
        return DMSFile(self, file_mode)

    def __str__(self) -> str:
        return f"{self.fullname:<14} #{self.file_number:02d}  {self.low_core_addr:04o}  {self.entry_point:>04o}  {self.high_core_addr:>o}"

    def __repr__(self) -> str:
        return str(self)


class DirectorNameBlock:
    """
    Directory Name (DN) Block

    https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf Pag 102

        +-----------------------------------+
      0 | First scratch block number        | | First |  Disk = 0o373, DecTape = 0o005
      1 | 2-Digit version number            | | DN    |
      2 | First SAM block number            | | Only  |  = 0o200
        +-----------------------------------+
      3 | DN Entry 1                        |
        | ...                               |
        | DN Entry 25                       |
        +-----------------------------------+
        | Link to next directory name       |
        +-----------------------------------+
    """

    fs: "DMSFilesystem"
    # Block number
    block_number: int = 0
    # Block sequence number
    block_seq_nr: int = 0
    # First scratch block number
    first_scratch_block_number: int = 0
    # Version number
    version_number: int = 0
    # First SAM block number
    first_sam_block_number: int = 0
    # Link to next directory name block number
    next_directory_name_block: int = 0
    # Directory entries
    entries: t.Dict[int, "DMSDirectoryEntry"]  # File number => DMSDirectoryEntry

    def __init__(self, fs: "DMSFilesystem"):
        self.fs = fs
        self.entries = {}

    @classmethod
    def read(
        cls, fs: "DMSFilesystem", block_number: int, block_seq_nr: int, sam: t.Optional["StorageAllocationMap"] = None
    ) -> "DirectorNameBlock":
        """
        Read a Directory Name Block from disk
        """
        self = cls(fs)
        words = self.fs.read_words_block(block_number)
        self.block_number = block_number
        self.block_seq_nr = block_seq_nr
        self.first_scratch_block_number = words[0]
        self.version_number = words[1]
        self.first_sam_block_number = words[2]
        self.next_directory_name_block = words[3 + DN_ENTRIES * DN_ENTRY_SIZE]
        if sam is not None:
            for position in range(3, 3 + DN_ENTRIES * DN_ENTRY_SIZE, DN_ENTRY_SIZE):
                dir_entry = DMSDirectoryEntry.read(self, sam, words, position)
                if not dir_entry.is_empty:
                    self.entries[dir_entry.file_number] = dir_entry
        return self

    @classmethod
    def new(
        cls, fs: "DMSFilesystem", block_number: int, block_seq_nr: int, sam: t.Optional["StorageAllocationMap"] = None
    ) -> "DirectorNameBlock":
        """
        Create a new Directory Name Block
        """
        self = cls(fs)
        self.block_number = block_number
        self.block_seq_nr = block_seq_nr
        return self

    def write(self) -> None:
        """
        Write Directory Name Block
        """
        words = [
            self.first_scratch_block_number,
            self.version_number,
            self.first_sam_block_number,
        ]
        for file_number in range(self.block_seq_nr * DN_ENTRIES + 1, (self.block_seq_nr + 1) * DN_ENTRIES + 1):
            entry = self.entries.get(file_number)
            if entry is not None and not entry.is_empty:
                words.extend(entry.to_words())
            else:
                words.extend([0] * DN_ENTRY_SIZE)
        words.append(self.next_directory_name_block)
        assert len(words) == BLOCK_SIZE_WORDS
        self.fs.write_words_block(self.block_number, words)

    @property
    def first_file_number(self) -> int:
        """
        First file number in this segment
        """
        return self.block_seq_nr * DN_ENTRIES + 1

    @property
    def last_file_number(self) -> int:
        """
        Last file number in this segment
        """
        return min((self.block_seq_nr + 1) * DN_ENTRIES, MAX_FILE_NUMBER)

    @property
    def number_of_entries(self) -> int:
        """
        Number of directory entries in this segment
        """
        return len(self.entries_list)

    @property
    def entries_list(self) -> t.List["DMSDirectoryEntry"]:
        return [x for x in self.entries.values() if not x.is_empty]

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Directory Name Block\n")
        buf.write(f"Block number:          {self.block_number:>5}\n")
        buf.write(f"First scratch block:   {self.first_scratch_block_number:>5}\n")
        buf.write(f"Version number:        {self.version_number:>5}\n")
        buf.write(f"First SAM block:       {self.first_sam_block_number:>5}\n")
        buf.write(f"Next dir name block:   {self.next_directory_name_block:>5}\n")
        if self.entries_list:
            buf.write("\nFilename       Num  Low   Entry Core  Blocks")
            buf.write("\n                    Core  Point Bank")
            buf.write("\n--------       ---  ----  ----- ----  -----\n")
            for i, x in enumerate(self.entries.values()):
                buf.write(f"{x}     {x.get_blocks()}\n")
        return buf.getvalue()


class DMSFilesystem(AbstractFilesystem):
    """
    PDP-8 4k Disk Monitor System Filesystem

    https://svn.so-much-stuff.com/svn/trunk/pdp8/src/dec/dec-08-odsma/dec-08-odsma-a-d.pdf Pag 97

    - Blocks are 129 12-bit words (258 bytes)
    - Skip the first word of disk

    Directory Name (DN)
    Storage Allocation Map (SAM)

    Disk (pag 98)
    ====

    Block
           +---------------+
    0o177  |   DN1 (USER)  |
           +---------------+
    0o200  |   SAM1 (USER) |
           +---------------+
    0o201  |   DN2 (USER)  |
           +---------------+
    0o202  |   DN3 (USER)  |
           +---------------+
           |               |
    data   /               /
           |               |
           +---------------+
    0o373  |   Scratch     |
    0o374  |   Scratch     |
    0o375  |   Scratch     |
           +---------------+

    Dectape (Pag 100)
    =======

    Block

           +---------------+
    0o005  |   Scratch     |
    0o006  |   Scratch     |
    0o007  |   Scratch     |
           +---------------+
           /               /
           +---------------+
    0o177  |   DN1 (USER)  |
           +---------------+
    0o200  |   SAM1 (USER) |
           +---------------+
    0o201  |   DN2 (USER)  |
           +---------------+
    0o202  |   SAM2 (USER) |
           +---------------+
    0o203  |   SAM3 (USER) |
           +---------------+
    0o204  |   SAM4 (USER) |
           +---------------+
    0o205  |   SAM5 (USER) |
           +---------------+
    0o206  |   SAM6 (USER) |
           +---------------+
    0o207  |   DN3 (USER)  |
           +---------------+

    """

    fs_name = "dms"
    fs_description = "PDP-8 4k Disk Monitor System"
    fs_platforms = ["pdp-8"]
    dev: BlockDevice12Bit

    dectape: bool = False  # DECtape device
    version_string: str  # Version
    first_scratch_block_number: int  # First scratch block number
    first_sam_block_number: int  # First SAM block number

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"], device_type: t.Union[bool, str] = ""):
        if isinstance(file_or_device, AbstractFile):
            self.dev = RXBlockDevice12Bit(file_or_device)
            # Blocks are 129 12-bit words (258 bytes)
            self.dev.block_size_words = BLOCK_SIZE_WORDS
            self.dectape = device_type == DECTAPE
            if not self.dectape:
                # Skip the first word of disk
                self.dev.block_offset_bytes = BYTES_PER_WORD
        elif isinstance(file_or_device, BlockDevice12Bit):
            self.dev = file_or_device
            self.dectape = self.dev.block_offset_bytes == BYTES_PER_WORD
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        device_type: t.Union[bool, str] = "",
        **kwargs: t.Union[bool, str],
    ) -> "DMSFilesystem":
        """
        Mount the 4k Disk Monitor System filesystem from a file
        """
        self = cls(file_or_dev, device_type=device_type)
        # Read the first Directory Name block
        dn = DirectorNameBlock.read(self, DN_START, 0)
        if dn.first_sam_block_number != SAM_START:
            raise OSError(f"Invalid DMS filesystem: first SAM block number is not {SAM_START}")
        self.first_scratch_block_number = dn.first_scratch_block_number
        self.first_sam_block_number = dn.first_sam_block_number
        self.version_string = sixbit_word12_to_asc(dn.version_number)
        if strict:
            sam = StorageAllocationMap.read(self)
            reserved_blocks = sam.files_blocks.get(RESERVED_FILE_NUMBER, [])
            if not reserved_blocks or not self.first_scratch_block_number in reserved_blocks:
                raise OSError(errno.EIO, os.strerror(errno.EIO))
        return self

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as 129 12bit words

        - Blocks are 129 12-bit words (258 bytes)
        - Skip the first word of disk
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(self, block_number: int, words: t.List[int]) -> None:
        """
        Write a block as 129 12bit words

        - Blocks are 129 12-bit words (258 bytes)
        - Skip the first word of disk
        """
        return self.dev.write_words_block(block_number, words)

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
        sam: t.Optional["StorageAllocationMap"] = None,
    ) -> t.Iterator["DMSDirectoryEntry"]:
        dms_filename = DMSFilename(pattern, wildcard=wildcard) if pattern else None
        for dn in self.read_directory_name_blocks(sam):
            for entry in dn.entries_list:
                if dms_filename is None or dms_filename.match(entry):
                    yield entry

    @property
    def entries_list(self) -> t.Iterator["DMSDirectoryEntry"]:
        for dn in self.read_directory_name_blocks():
            for entry in dn.entries_list:
                yield entry

    def get_file_entry(self, fullname: str) -> DMSDirectoryEntry:
        """
        Get the directory entry for a file
        """
        dms_filename = DMSFilename(fullname)
        for dn in self.read_directory_name_blocks():
            for entry in dn.entries_list:
                if (
                    entry.filename.strip() == dms_filename.filename.strip()
                    and entry.program_type == dms_filename.program_type
                ):
                    return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def read_bytes(self, fullname: str, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of a file"""
        if file_mode is None:
            dms_filename = DMSFilename(fullname)
            if dms_filename.program_type == FILE_TYPE_ASCII:
                file_mode = ASCII
            else:
                file_mode = IMAGE
        # Always read the file as IMAGE
        f = self.open_file(fullname, IMAGE)
        try:
            data = f.read_block(0, READ_FILE_FULL)
            if file_mode == ASCII:
                # Convert IMAGE => words => ASCII
                words = from_bytes_to_12bit_words(data, file_mode=IMAGE)
                return from_12bit_words_to_bytes(words, file_mode=ASCII)
            else:
                return data
        finally:
            f.close()

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        dms_filename = DMSFilename(fullname)
        if file_mode is None:
            if dms_filename.program_type == FILE_TYPE_ASCII:
                file_mode = ASCII
            else:
                file_mode = IMAGE
        if file_mode == ASCII:
            # Append FF (form feed) if not present
            if not content.endswith(b'\x0c\x0c'):
                content = content + b'\x0c\x0c'
            # Convert ASCII => words
            words = from_bytes_to_12bit_words(content, file_mode=ASCII)
        else:
            # Convert IMAGE => words
            words = from_bytes_to_12bit_words(content, file_mode=IMAGE)
        # Allocate space
        number_of_blocks = int(math.ceil(len(words) * 1.0 / DATA_BLOCK_SIZE_WORDS))
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        # Write blocks
        if entry is not None:
            blocks = entry.get_blocks()
            for i, block in enumerate(blocks):
                block_words = words[i * DATA_BLOCK_SIZE_WORDS : (i + 1) * DATA_BLOCK_SIZE_WORDS]
                block_words.extend([0] * (BLOCK_SIZE_WORDS - len(block_words)))
                # The last word of the block is the link to the next block
                block_words[-1] = blocks[i + 1] if i + 1 < len(blocks) else 0
                entry.dn.fs.write_words_block(block, block_words)

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[DMSDirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        # Delete the file if it already exists
        dms_filename = DMSFilename(fullname)
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        # Allocate space
        sam = StorageAllocationMap.read(self)
        file_number = sam.allocate_space(fullname, number_of_blocks)
        # Create entry
        entry = None
        for dn in self.read_directory_name_blocks():
            if dn.first_file_number <= file_number <= dn.last_file_number:
                entry = DMSDirectoryEntry(dn, sam)
                entry.filename = dms_filename.filename
                entry.file_number = file_number
                entry.program_type = dms_filename.program_type
                entry.system_program = dms_filename.system_program
                entry.high_core_addr = (dms_filename.core_addr >> 12) & 0o7
                entry.low_core_addr = dms_filename.core_addr & 0o7777
                entry.entry_point = dms_filename.entry_point
                dn.entries[file_number] = entry
                dn.write()
                break
        if entry is not None:
            sam.write()
        return entry

    def isdir(self, fullname: str) -> bool:
        return False

    def chdir(self, fullname: str) -> bool:
        return False

    def read_directory_name_blocks(
        self, sam: t.Optional["StorageAllocationMap"] = None
    ) -> t.Iterator["DirectorNameBlock"]:
        """
        Read the Directory Name Blocks
        """
        if sam is None:
            sam = StorageAllocationMap.read(self)
        next_block_number = DN_START
        block_seq_nr = 0
        while next_block_number != 0:
            dn = DirectorNameBlock.read(self, next_block_number, block_seq_nr, sam)
            next_block_number = dn.next_directory_name_block
            block_seq_nr += 1
            yield dn

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return StorageAllocationMap.read(self).free()

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        sam = StorageAllocationMap.read(self)
        if not options.get("brief"):
            # Number of free blocks
            sys.stdout.write(f"\nFB={sam.free():>04o}\n")
            sys.stdout.write("\n")
            sys.stdout.write("NAME  TYPE    BLK\n")
            sys.stdout.write("\n")
            sys.stdout.write(f"{self.version_string}\n")
        for x in self.filter_entries_list(pattern, include_all=True, sam=sam):
            if x.file_number == RESERVED_FILE_NUMBER and not options.get("full"):
                continue
            if options.get("brief"):
                # Lists only file name and extension
                sys.stdout.write(f"{x.filename:<4}.{x.extension}\n")
            else:
                # Filename, extension, core bank (for SYS/USER files) and length in blocks (in octal)
                if x.program_type == FILE_TYPE_SYS_USER:
                    fullname = f"{x.filename:<4}.{x.extension:<4}({x.high_core_addr:>o})"
                else:
                    fullname = f"{x.filename:<4}.{x.extension:<7}"
                sys.stdout.write(f"{fullname} {x.get_length():>04o}\n")
        sys.stdout.write("\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            # Display the SAM
            sam = StorageAllocationMap.read(self)
            for sam_block in sam.sam_blocks:
                print(sam_block)
        elif arg:
            sam = StorageAllocationMap.read(self)
            sys.stdout.write("Filename       Num  Low   Entry Core\n")
            sys.stdout.write("                    Core  Point Bank\n")
            sys.stdout.write("--------       ---  ----  ----- ----\n")
            for entry in self.filter_entries_list(arg, include_all=True, sam=sam):
                sys.stdout.write(f"{entry}\n")
        else:
            sam = StorageAllocationMap.read(self)
            sam_blocks = [x.block_number for x in sam.sam_blocks]
            sys.stdout.write(f"SAM blocks:               {sam_blocks}\n")
            sys.stdout.write(f"Is DECtape:               {'YES' if self.dectape else 'NO'}\n")
            for dn in self.read_directory_name_blocks():
                sys.stdout.write(f"{dn}\n")

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            entry = self.get_file_entry(fullname)
            if start is None:
                start = 0
            if end is None:
                end = entry.get_length() - 1
            blocks = entry.get_blocks()
            for block_number in range(start, end + 1):
                words = self.read_words_block(blocks[block_number])
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = self.get_size() // BLOCK_SIZE_WORDS // BYTES_PER_WORD - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)

    @classmethod
    def initialize(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        device_type: t.Union[bool, str] = "",
        **kwargs: t.Union[bool, str],
    ) -> "DMSFilesystem":
        """
        Create an empty PDP-8 4k Disk Monitor filesystem
        """
        self = cls(file_or_dev, device_type=device_type)
        version_string = "AF"
        if self.dectape:
            dn_blocks = DECTAPE_DN_BLOCKS
            sam_blocks = DECTAPE_SAM_BLOCKS
            scratch_blocks = DECTAPE_SCRATCH_BLOCKS
        else:
            dn_blocks = DN_BLOCKS
            sam_blocks = SAM_BLOCK
            scratch_blocks = SCRATCH_BLOCKS
        # Initialize this instance
        self.first_scratch_block_number = scratch_blocks[0]
        self.first_sam_block_number = sam_blocks[0]
        self.version_string = version_string
        # Create SAM and allocate blocks
        sam = StorageAllocationMap.new(self, sam_blocks)
        for block_number in scratch_blocks + dn_blocks + sam_blocks:
            sam.set_block(block_number, RESERVED_FILE_NUMBER)
        sam.write()
        # Create DNs
        for i, block_number in enumerate(dn_blocks):
            dn = DirectorNameBlock.new(self, block_number, i)
            if i == 0:
                dn.first_scratch_block_number = scratch_blocks[0]
                dn.first_sam_block_number = sam_blocks[0]
                dn.version_number = asc_to_sixbit_word12(version_string)
                entry = DMSDirectoryEntry(dn, sam)
                entry.filename = MONITOR_FILENAME
                entry.file_number = RESERVED_FILE_NUMBER
                entry.program_type = FILE_TYPE_SYS_USER
                entry.system_program = True
                entry.high_core_addr = 0
                entry.low_core_addr = 0o7000
                entry.entry_point = 0o7000
                dn.entries[entry.file_number] = entry
            if i < len(dn_blocks) - 1:
                dn.next_directory_name_block = dn_blocks[i + 1]
            dn.write()
        sam.write()
        return self

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def close(self) -> None:
        self.dev.close()

    def get_pwd(self) -> str:
        return ""

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return [
            EXT_SYS,
            EXT_USER,
            EXT_ASCII,
            EXT_BINARY,
            EXT_FTC_BIN,
        ]
